"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import numpy as np
import jax
import jax.numpy as jnp

from minimax.envs.registration import register_comparator


@jax.jit
def is_equal_map(a, b):
	agent_pos_eq = jnp.equal(a.agent_pos, b.agent_pos).all()
	goal_pos_eq = jnp.equal(a.goal_pos, b.goal_pos).all()
	wall_map_eq = jnp.equal(a.wall_map, b.wall_map).all()

	_eq = jnp.logical_and(agent_pos_eq, goal_pos_eq)
	_eq = jnp.logical_and(_eq, wall_map_eq)

	return _eq


# Register the mutators
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register_comparator(env_id='Maze', comparator_id=None, entry_point=module_path + ':is_equal_map')


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file is modified from 
https://github.com/openai/baselines

Licensed under the MIT License;
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://opensource.org/license/mit/
"""


import os
import sys
import json
import csv
import time
import datetime
import copy
import logging
import git
from typing import Dict

import minimax.util.checkpoint as _checkpoint_util


class KVWriter(object):
	def writekvs(self, kvs):
		raise NotImplementedError

class SeqWriter(object):
	def writeseq(self, seq):
		raise NotImplementedError

class HumanOutputFormat(KVWriter, SeqWriter):
	def __init__(self, filename_or_file):
		if isinstance(filename_or_file, str):
			self.file = open(filename_or_file, 'wt')
			self.own_file = True
		else:
			assert hasattr(filename_or_file, 'read'), 'expected file or str, got %s'%filename_or_file
			self.file = filename_or_file
			self.own_file = False

	def writekvs(self, kvs):
		# Create strings for printing
		key2str = {}
		for (key, val) in sorted(kvs.items()):
			if hasattr(val, '__float__'):
				valstr = '%-8.3g' % val
			else:
				valstr = str(val)
			key2str[self._truncate(key)] = self._truncate(valstr)

		# Find max widths
		if len(key2str) == 0:
			print('WARNING: tried to write empty key-value dict')
			return
		else:
			keywidth = max(map(len, key2str.keys()))
			valwidth = max(map(len, key2str.values()))

		# Write out the data
		dashes = '-' * (keywidth + valwidth + 7)
		lines = [dashes]
		for (key, val) in sorted(key2str.items(), key=lambda kv: kv[0].lower()):
			lines.append('| %s%s | %s%s |' % (
				key,
				' ' * (keywidth - len(key)),
				val,
				' ' * (valwidth - len(val)),
			))
		lines.append(dashes)
		self.file.write('\n'.join(lines) + '\n')

		# Flush the output to the file
		self.file.flush()

	def _truncate(self, s):
		maxlen = 64
		return s[:maxlen-3] + '...' if len(s) > maxlen else s

	def writeseq(self, seq):
		seq = list(seq)
		for (i, elem) in enumerate(seq):
			self.file.write(elem)
			if i < len(seq) - 1: # add space unless this is the last one
				self.file.write(' ')
		self.file.write('\n')
		self.file.flush()

	def close(self):
		if self.own_file:
			self.file.close()


def gather_metadata() -> Dict:
	date_start = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
	# Gathering git metadata.
	try:
		import git

		try:
			repo = git.Repo(search_parent_directories=True)
			git_sha = repo.commit().hexsha
			git_data = dict(
				commit=git_sha,
				branch=None if repo.head.is_detached else repo.active_branch.name,
				is_dirty=repo.is_dirty(),
				path=repo.git_dir,
			)
		except git.InvalidGitRepositoryError:
			git_data = None
	except ImportError:
		git_data = None

	# Gathering slurm metadata.
	if "SLURM_JOB_ID" in os.environ:
		slurm_env_keys = [k for k in os.environ if k.startswith("SLURM")]
		slurm_data = {}
		for k in slurm_env_keys:
			d_key = k.replace("SLURM_", "").replace("SLURMD_", "").lower()
			slurm_data[d_key] = os.environ[k]
	else:
		slurm_data = None
	return dict(
		date_start=date_start,
		date_end=None,
		successful=False,
		git=git_data,
		slurm=slurm_data,
		env=os.environ.copy(),
	)


class Logger:
	def __init__(
		self,
		log_dir='~/logs/minimax',
		xpid=None,
		xp_args=None,
		callback=None,
		from_last_checkpoint=False,
		verbose=False):
		# Set up checkpoint meta
		self.verbose = verbose
		if self.verbose:
			self._stdout = HumanOutputFormat(sys.stdout)

		self._callback = callback

		formatter = logging.Formatter("%(message)s")
		self._logger = logging.getLogger("logs/out")
		shandle = logging.StreamHandler()
		shandle.setFormatter(formatter)
		self._logger.addHandler(shandle)
		self._logger.setLevel(logging.INFO)

		# Set up main paths for logs and checkpoints
		self.paths = {}
		log_dir_path = os.path.expandvars(os.path.expanduser(log_dir))
		xpid_dir_path = os.path.join(log_dir_path, xpid)
		if not xpid:
			xpid = "{proc}_{unixtime}".format(
				proc=os.getpid(), unixtime=int(time.time())
			)
		if not os.path.exists(xpid_dir_path):
			self._logger.info("Creating log directory: %s", xpid_dir_path)
			os.makedirs(xpid_dir_path, exist_ok=True)
		self.paths['log_dir'] = log_dir_path
		self.paths['xpid_dir'] = xpid_dir_path
		self.paths['checkpoint'] = os.path.join(xpid_dir_path, 'checkpoint.pkl')

		# Create logs.csv file
		logs_csv_path = os.path.join(xpid_dir_path, 'logs.csv')
		self.paths['log_csv'] = logs_csv_path
		self._last_n_logged_lines = 0
		self._last_logged_tick = self._get_last_logged_tick()

		self.append_to_existing_logs = \
			self._last_logged_tick >= 0 \
			and from_last_checkpoint \
			and os.path.exists(self.paths['checkpoint'])
		log_mode = "a" if self.append_to_existing_logs else "w+"
		self._logfile = open(logs_csv_path, log_mode)
		self._logwriter = None

		# Create meta file
		if xp_args is not None:
			meta_path = os.path.join(xpid_dir_path, 'meta.json')

			meta = gather_metadata()
			meta['config'] = dict(xp_args)
			meta['xpid'] = xpid

			self._save_metadata(meta_path, meta)

	def _save_metadata(self, meta_path, meta):
		with open(meta_path, "w") as jsonfile:
			json.dump(meta, jsonfile, indent=4, sort_keys=True)

	def _get_last_logged_tick(self):
		last_tick = -1
		logs_csv_path = self.paths['log_csv']
		if os.path.exists(logs_csv_path):
			with open(logs_csv_path, "r") as csvfile:
				reader = csv.reader(csvfile)
				lines = list(reader)
				# Need at least two lines in order to read the last tick:
				# the first is the csv header and the second is the first line
				# of data.
				if len(lines) > 1:
					self._last_n_logged_lines = len(lines)
					try:
						last_tick = int(lines[-1][0])
					except:
						last_tick = -1

		return last_tick

	def log(self, stats, _tick, ignore_val=None):
		if ignore_val is not None:
			stats = {k:v if v != ignore_val else None for k,v in stats.items()}

		_stats = {
			'_tick': _tick,
			'_time': time.time()
		}
		_stats.update(stats)
		stats = _stats

		if self._logwriter is None:
			fieldnames = list(stats.keys())
			self._logwriter = csv.DictWriter(self._logfile, fieldnames=fieldnames)

		if _tick > self._last_logged_tick \
			or not self.append_to_existing_logs:
			if self._last_n_logged_lines == 0:
				fieldnames = list(stats.keys())
				self._logfile.write("# %s\n" % ",".join(fieldnames))
				self._logfile.flush()
				self._last_n_logged_lines = 1

			self._logwriter.writerow(stats)
			self._logfile.flush()

			if self._callback is not None:
				self._callback(stats)

		if self.verbose:
			self._stdout.writekvs(stats)

	@property
	def checkpoint_path(self):
		return self.paths['checkpoint']

	def checkpoint(
		self, 
		runner_state, 
		name='checkpoint', 
		index=None,
		archive_interval=None):
		_checkpoint_util.safe_checkpoint(
			runner_state,
			self.paths['xpid_dir'],
			name,
			index,
			archive_interval
		)

	def load_last_checkpoint_state(self):
		checkpoint_path = \
			os.path.join(self.paths['xpid_dir'], f'checkpoint.pkl')

		if os.path.exists(checkpoint_path):
			self._logger.info(f'Loading previous checkpoint from {checkpoint_path}...')
			return _checkpoint_util.load_pkl_object(checkpoint_path)
		else:
			return None



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from enum import Enum
from typing import Tuple, Optional

import numpy as np
import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P
import optax
import flax
import flax.linen as nn
from flax.core.frozen_dict import FrozenDict

import minimax.envs as envs
from minimax.runners.dr_runner import DRRunner
from minimax.util import pytree as _tree_util
from minimax.util.rl import (
	AgentPop,
	VmapTrainState,
	RolloutStorage,
	RollingStats,
	UEDScore,
	compute_ued_scores,
	PopPLRManager
)


class MutationCriterion(Enum):
	BATCH = 0
	EASY = 1
	HARD = 2


class PLRRunner(DRRunner):
	def __init__(
		self, 
		*,
		replay_prob=0.5,
		buffer_size=100,
		staleness_coef=0.3,
		use_score_ranks=True,
		temp=1.0,
		min_fill_ratio=0.5,
		use_robust_plr=False,
		use_parallel_eval=False,
		ued_score='l1_value_loss',
		force_unique=False, # Slower if True
		mutation_fn=None,
		n_mutations=0,
		mutation_criterion='batch',
		mutation_subsample_size=1,
		**kwargs):
		use_mutations = mutation_fn is not None
		if use_parallel_eval:
			replay_prob = 1.0 # Replay every rollout cycle
			mutation_criterion = 'batch' # Force batch mutations (no UED scores)
			self._n_parallel_batches = 3 if use_mutations else 2
			kwargs['n_parallel'] *= self._n_parallel_batches

		super().__init__(**kwargs)

		self.replay_prob = replay_prob
		self.buffer_size = buffer_size
		self.staleness_coef = staleness_coef
		self.temp = temp
		self.use_score_ranks = use_score_ranks
		self.min_fill_ratio = min_fill_ratio
		self.use_robust_plr = use_robust_plr
		self.use_parallel_eval = use_parallel_eval
		self.ued_score = UEDScore[ued_score.upper()]

		self.use_mutations = use_mutations
		if self.use_mutations:
			self.mutation_fn = envs.get_mutator(self.benv.env_name, mutation_fn)
		else:
			self.mutation_fn = None
		self.n_mutations = n_mutations
		self.mutation_criterion = MutationCriterion[mutation_criterion.upper()]
		self.mutation_subsample_size = mutation_subsample_size

		self.force_unique = force_unique
		if force_unique:
			self.comparator_fn = envs.get_comparator(self.benv.env_name)
		else:
			self.comparator_fn = None

		if mutation_fn is not None and mutation_criterion != 'batch':
			assert self.n_parallel % self.mutation_subsample_size == 0, \
				'Number of parallel envs must be divisible by mutation subsample size.'

	def reset(self, rng):
		runner_state = list(super().reset(rng))
		rng = runner_state[0]
		runner_state[0], subrng = jax.random.split(rng)
		example_state = self.benv.env.reset(rng)[1]

		self.plr_mgr = PopPLRManager(
			n_agents=self.n_students,
			example_level=example_state,
			ued_score=self.ued_score,
			replay_prob=self.replay_prob,
			buffer_size=self.buffer_size,
			staleness_coef=self.staleness_coef,
			temp=self.temp,
			use_score_ranks=self.use_score_ranks,
			min_fill_ratio=self.min_fill_ratio,
			use_robust_plr=self.use_robust_plr,
			use_parallel_eval=self.use_parallel_eval,
			comparator_fn=self.comparator_fn,
			n_devices=self.n_devices
		)
		plr_buffer = self.plr_mgr.reset(self.n_students)

		train_state = runner_state[1]
		train_state = train_state.replace(plr_buffer=plr_buffer)
		if self.n_devices == 1:
			runner_state[1] = train_state
		else:
			plr_buffer = jax.tree_map(lambda x: x.repeat(self.n_devices, 1), plr_buffer) # replicate plr buffer
			runner_state += (plr_buffer,) # Return PLR buffer directly to make shmap easier

		self.dummy_eval_output = self._create_dummy_eval_output(train_state)

		return tuple(runner_state)

	def _create_dummy_eval_output(self, train_state):
		rng, *vrngs = jax.random.split(jax.random.PRNGKey(0), self.n_students+1)
		obs, state, extra = self.benv.reset(jnp.array(vrngs))

		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=(self.n_students, self.n_parallel*self.n_eval))

		ued_scores = jnp.zeros((self.n_students, self.n_parallel))

		if self.student_pop.agent.is_recurrent:
			carry = self.zero_carry
		else:
			carry = None
		rollout = self.student_rollout.reset()

		train_batch = self.student_rollout.get_batch(
			rollout, 
			self.student_pop.get_value(
				jax.lax.stop_gradient(train_state.params), 
				obs, 
				carry,
			)
		)

		return (
			rng,
			train_state, 
			state, 
			state,
			obs, 
			carry, 
			extra, 
			ep_stats,
			state,
			train_batch,
			ued_scores
		)

	@partial(jax.jit, static_argnums=(0,8))
	def _eval_and_update_plr(
			self,
			rng,
			levels,
			level_idxs, 
			train_state,    
			update_plr,
			parent_idxs=None,
			dupe_mask=None,
			fake=False):
		# Collect rollout and optionally update plr buffer
		# Returns train_batch and ued_scores
		# Perform rollout: @todo: pmap this
		if fake:
			dummy_eval_output = list(self.dummy_eval_output)
			dummy_eval_output[1] = train_state
			return tuple(dummy_eval_output)

		rollout_batch_shape = (self.n_students, self.n_parallel*self.n_eval)
		obs, state, extra = self.benv.set_state(levels)
		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=rollout_batch_shape)

		rollout_start_state = state

		done = jnp.zeros(rollout_batch_shape, dtype=jnp.bool_)
		if self.student_pop.agent.is_recurrent:
			carry = self.zero_carry
		else:
			carry = None

		rng, subrng = jax.random.split(rng)
		start_state = state
		reset_state = state
		rollout, state, start_state, obs, carry, extra, ep_stats, train_state = \
			self._rollout_students(
				subrng, 
				train_state, 
				state, 
				start_state,
				obs, 
				carry, 
				done,
				reset_state,
				extra, 
				ep_stats
			)

		train_batch = self.student_rollout.get_batch(
			rollout, 
			self.student_pop.get_value(
				jax.lax.stop_gradient(train_state.params), 
				obs, 
				carry
			)
		)

		# Update PLR buffer
		if self.ued_score == UEDScore.MAX_MC:
			max_returns = jax.vmap(lambda x,y: x.at[y].get())(train_state.plr_buffer.max_returns, level_idxs)
			max_returns = jnp.where(
				jnp.greater_equal(level_idxs, 0),
				max_returns,
				jnp.full_like(max_returns, -jnp.inf)
			)
			ued_info = {'max_returns': max_returns}
		else:
			ued_info = None
		ued_scores, ued_score_info = compute_ued_scores(
			self.ued_score, train_batch, self.n_eval, info=ued_info, ignore_val=-jnp.inf, per_agent=True)
		next_plr_buffer = self.plr_mgr.update(
			train_state.plr_buffer, 
			levels=levels, 
			level_idxs=level_idxs, 
			ued_scores=ued_scores,
			dupe_mask=dupe_mask, 
			info=ued_score_info, 
			ignore_val=-jnp.inf,
			parent_idxs=parent_idxs)

		next_plr_buffer = jax.vmap(
			lambda update, new, prev: jax.tree_map(
				lambda x, y: jax.lax.select(update, x, y), new, prev)
		)(update_plr, next_plr_buffer, train_state.plr_buffer)

		train_state = train_state.replace(plr_buffer=next_plr_buffer)

		return (
			rng,
			train_state, 
			state, 
			start_state, 
			obs, 
			carry, 
			extra, 
			ep_stats,
			rollout_start_state,
			train_batch,
			ued_scores,
		)

	@partial(jax.jit, static_argnums=(0,))
	def _mutate_levels(self, rng, levels, level_idxs, ued_scores=None):
		if not self.use_mutations:
			return levels, level_idxs, jnp.full_like(level_idxs, -1)

		def upsample_levels(levels, level_idxs, subsample_idxs):
			subsample_idxs = subsample_idxs.repeat(self.n_parallel//self.mutation_subsample_size, -1)
			parent_idxs = level_idxs.take(subsample_idxs)
			levels = jax.vmap(
				lambda x, y: jax.tree_map(lambda _x: jnp.array(_x).take(y, 0), x)
			)(levels, parent_idxs)
			
			return levels, parent_idxs

		if self.mutation_criterion == MutationCriterion.BATCH:
			parent_idxs = level_idxs

		if self.mutation_criterion == MutationCriterion.EASY:
			_, top_level_idxs = jax.lax.approx_min_k(ued_scores, self.mutation_subsample_size)
			levels, parent_idxs = upsample_levels(levels, level_idxs, top_level_idxs)

		elif self.mutation_criterion == MutationCriterion.HARD:
			_, top_level_idxs = jax.lax.approx_max_k(ued_scores, self.mutation_subsample_size)
			levels, parent_idxs = upsample_levels(levels, level_idxs, top_level_idxs)

		n_parallel = level_idxs.shape[-1]
		vrngs = jax.vmap(lambda subrng: jax.random.split(subrng, n_parallel))(
			jax.random.split(rng, self.n_students)
		)

		mutated_levels = jax.vmap(
			lambda *args: jax.vmap(self.mutation_fn, in_axes=(0,None,0,None))(*args),
			in_axes=(0,None,0,None)
		)(vrngs, self.benv.env_params, levels, self.n_mutations)

		# Mutated levels do not have existing idxs in the PLR buffer.
		mutated_level_idxs = jnp.full((self.n_students, n_parallel), -1)

		return mutated_levels, mutated_level_idxs, parent_idxs

	def _efficient_grad_update(self, rng, train_state, train_batch, is_replay):
		# PPOAgent vmaps over the train state and batch. Batch must be N x EM
		skip_grad_update = jnp.logical_and(self.use_robust_plr, ~is_replay)

		if self.n_students == 1:
			train_state, stats = jax.lax.cond(
				skip_grad_update[0],
				partial(self.student_pop.update, fake=True),
				self.student_pop.update,
				*(rng, train_state, train_batch)
			)
		elif self.n_students > 1: # Have to vmap all students + take only students that need updates
			_, dummy_stats = jax.vmap(lambda *_: self.student_pop.agent.get_empty_update_stats())(np.arange(self.n_students))
			_train_state, stats = self.student.update(rng, train_state, train_batch)
			train_state, stats = jax.vmap(lambda cond,x,y: \
					jax.tree_map(lambda _cond,_x,_y: jax.lax.select(_cond,_x,_y), cond, x, y))(
						is_replay, (train_state, stats), (_train_state, dummy_stats)
					)

		return train_state, stats

	@partial(jax.jit, static_argnums=(0,))
	def _compile_stats(self, update_stats, ep_stats, is_replay, env_metrics=None, plr_stats=None):
		stats = super()._compile_stats(update_stats, ep_stats, env_metrics, mask_passable=False)

		if self.track_env_metrics:
			if self.use_parallel_eval:
				env_passable_mask = env_metrics.pop('env/passable')
				env_env_metrics = {k:v for k,v in env_metrics.items() if k.startswith('env/')}
				env_env_metrics = jax.vmap(lambda info, mask: jax.tree.map(lambda x: (x*mask).sum()/mask.sum(), info))(env_env_metrics, env_passable_mask)
				env_env_metrics.update({'env/passable_ratio': jax.vmap(lambda x: x.mean())(env_passable_mask)})
				env_metrics.update(env_env_metrics)

				plr_passable_mask = env_metrics.pop('plr/passable')
				plr_env_metrics = {k:v for k,v in env_metrics.items() if k.startswith('plr/')}
				plr_env_metrics = jax.vmap(lambda info, mask: jax.tree.map(lambda x: (x*mask).sum()/mask.sum(), info))(plr_env_metrics, plr_passable_mask)
				plr_env_metrics.update({'plr/passable_ratio': jax.vmap(lambda x: x.mean())(plr_passable_mask)})
				env_metrics.update(plr_env_metrics)
			else:
				passable_mask = env_metrics.pop('passable')
				env_metrics = jax.vmap(lambda info, mask: jax.tree.map(lambda x: (x*mask).sum()/mask.sum(), info))(env_metrics, passable_mask)
				env_metrics.update({'passable_ratio': jax.vmap(lambda x: x.mean())(passable_mask)})

			if not self.use_parallel_eval:
				_env_metrics = jax.vmap(lambda cond: jax.lax.cond(
					cond,
					lambda *_: {f'env/{k}':-jnp.inf for k,_ in env_metrics.items()},
					lambda *_: {f'env/{k}':v.mean() for k,v in env_metrics.items()}
				))(is_replay)

				plr_env_metrics = jax.vmap(lambda cond: jax.lax.cond(
					cond,
					lambda *_: {f'plr/{k}':v.mean() for k,v in env_metrics.items()},
					lambda *_: {f'plr/{k}':-jnp.inf for k,_ in env_metrics.items()},
				))(is_replay)

				stats.update(jax.tree.map(lambda x: x[0], plr_env_metrics))
				stats.update(jax.tree.map(lambda x: x[0], _env_metrics))
			else:
				stats = {k:v for k,v in stats.items() if not k.startswith('env/')}
				env_metrics = {k:v.mean() for k,v in env_metrics.items()}
				stats.update(env_metrics)

		if plr_stats is not None:
			plr_stats = jax.vmap(lambda info: jax.tree_util.tree_map(lambda x: x.mean(), info))(plr_stats)
			plr_stats = jax.tree.map(lambda x: x[0], plr_stats)
			stats.update({f'plr/{k}':v for k,v in plr_stats.items()})

		if self.n_devices > 1:
			stats = jax.tree.map(lambda x: jax.lax.pmean(x, 'device'), stats)

		return stats


	@partial(jax.jit, static_argnums=(0,))
	def run(
		self, 
		rng, 
		train_state, 
		state, 
		start_state,
		obs, 
		carry=None, 
		extra=None, 
		ep_stats=None,
		plr_buffer=None):
		# If device sharded, load sharded PLR buffer into train state
		if self.n_devices > 1:
			rng = jax.random.fold_in(rng, jax.lax.axis_index('device'))
			train_state = train_state.replace(plr_buffer=plr_buffer)

		# Sample next training levels via PLR
		rng, *vrngs = jax.random.split(rng, self.n_students+1)
		obs, state, extra = self.benv.reset(jnp.array(vrngs), self.n_parallel, 1)

		if self.use_parallel_eval:
			n_level_samples = self.n_parallel//self._n_parallel_batches
			new_levels = jax.tree_map(lambda x: x.at[:,n_level_samples:2*n_level_samples].get(), state)
		else:
			n_level_samples = self.n_parallel
			new_levels = state

		rng, subrng = jax.random.split(rng)
		levels, level_idxs, is_replay, next_plr_buffer = \
			self.plr_mgr.sample(subrng, train_state.plr_buffer, new_levels, n_level_samples)
		train_state = train_state.replace(plr_buffer=next_plr_buffer)

		# If use_parallel_eval=True, need to combine replay and non-replay levels together
		# Need to mutate levels as well
		parent_idxs = jnp.full((self.n_students, self.n_parallel), -1)
		if self.use_parallel_eval: # Parallel ACCEL
			new_level_idxs = jnp.full_like(parent_idxs, -1)

			_all_levels = jax.vmap(
				lambda x,y: _tree_util.pytree_merge(x,y, start_idx=n_level_samples, src_len=n_level_samples),
				)(state, levels)
			_all_level_idxs = jax.vmap(
				lambda x,y: _tree_util.pytree_merge(x,y, start_idx=n_level_samples, src_len=n_level_samples),
				)(new_level_idxs, level_idxs)

			if self.use_mutations:
				rng, subrng = jax.random.split(rng)
				mutated_levels, mutated_level_idxs, _parent_idxs = self._mutate_levels(subrng, levels, level_idxs)
				
				fallback_levels = jax.tree_map(lambda x: x.at[:,-n_level_samples:].get(), state)
				fallback_level_idxs = jnp.full_like(mutated_level_idxs, -1)

				mutated_levels = jax.vmap(
					lambda cond,x,y: jax.tree_map(
						lambda _x,_y: jax.lax.select(cond,_x,_y), x, y
					))(is_replay, mutated_levels, fallback_levels)

				mutated_level_idxs = jax.vmap(
					lambda cond,x,y: jax.tree_map(
						lambda _x,_y: jax.lax.select(cond,_x,_y), x, y
					))(is_replay, mutated_level_idxs, fallback_level_idxs)

				_parent_idxs = jax.vmap(
					lambda cond,x,y: jax.tree_map(
						lambda _x,_y: jax.lax.select(cond,_x,_y), x, y
					))(is_replay, _parent_idxs, fallback_level_idxs)
		
				mutated_levels_start_idx = 2*n_level_samples
				_all_levels = jax.vmap(
					lambda x,y: _tree_util.pytree_merge(x,y, start_idx=mutated_levels_start_idx, src_len=n_level_samples),
					)(_all_levels, mutated_levels)
				_all_level_idxs = jax.vmap(
					lambda x,y: _tree_util.pytree_merge(x,y, start_idx=mutated_levels_start_idx, src_len=n_level_samples),
					)(_all_level_idxs, mutated_level_idxs)
				parent_idxs = jax.vmap(
					lambda x,y: _tree_util.pytree_merge(x,y, start_idx=mutated_levels_start_idx, src_len=n_level_samples),
					)(parent_idxs, _parent_idxs)

			levels = _all_levels
			level_idxs = _all_level_idxs

		# dedupe levels, move into PLR buffer logic
		if self.force_unique:
			level_idxs, dupe_mask = self.plr_mgr.dedupe_levels(next_plr_buffer, levels, level_idxs)
		else:
			dupe_mask = None 

		# Evaluate levels + update PLR
		result = self._eval_and_update_plr(
			rng, levels, level_idxs, train_state, update_plr=jnp.array([True]*self.n_students), parent_idxs=parent_idxs, dupe_mask=dupe_mask)
		rng, train_state, state, start_state, obs, carry, extra, ep_stats, \
			rollout_start_state, train_batch, ued_scores = result

		if self.use_parallel_eval:
			replay_start_idx = self.n_eval*n_level_samples
			replay_end_idx = 2*replay_start_idx
			train_batch = jax.vmap(
				lambda x: jax.tree_map(
					lambda _x: _x.at[:,replay_start_idx:replay_end_idx].get(), x)
				)(train_batch)

		# Gradient update
		rng, subrng = jax.random.split(rng)
		train_state, update_stats = self._efficient_grad_update(subrng, train_state, train_batch, is_replay)

		# Mutation step
		use_mutations = jnp.logical_and(self.use_mutations, is_replay)
		use_mutations = jnp.logical_and(use_mutations, not self.use_parallel_eval) # Already mutated above in parallel
		rng, arng, brng = jax.random.split(rng, 3)

		mutated_levels, mutated_level_idxs, parent_idxs = jax.lax.cond(
			use_mutations.any(),
			self._mutate_levels,
			lambda *_: (levels, level_idxs, jnp.full_like(level_idxs, -1)),
			*(arng, levels, level_idxs, ued_scores)
		)

		mutated_dupe_mask = jnp.zeros_like(mutated_level_idxs, dtype=jnp.bool_)
		if self.force_unique: # Should move into update plr logic
			mutated_level_idxs, mutated_dupe_mask = jax.lax.cond(
				use_mutations.any(),
				self.plr_mgr.dedupe_levels,
				lambda *_: (mutated_level_idxs, mutated_dupe_mask),
				*(next_plr_buffer, mutated_levels, mutated_level_idxs)
			)

		mutation_eval_result = jax.lax.cond(
			use_mutations.any(),
			self._eval_and_update_plr,
			partial(self._eval_and_update_plr, fake=True),
			*(brng, mutated_levels, mutated_level_idxs, train_state, use_mutations, parent_idxs, mutated_dupe_mask)
		)
		train_state = mutation_eval_result[1]

		# Collect training env metrics
		if self.track_env_metrics:
			if self.use_parallel_eval:
				new_levels = jax.tree.map(lambda x: x.at[:,:n_level_samples].get(), levels)
				replay_levels = jax.tree.map(lambda x: x.at[:,n_level_samples:2*n_level_samples].get(), levels)
				new_level_metrics = self.benv.get_env_metrics(new_levels)
				replay_level_metrics = self.benv.get_env_metrics(replay_levels)
				env_metrics = {f'env/{k}':v for k,v in new_level_metrics.items()}
				env_metrics.update({
						f'plr/{k}':v for k,v in replay_level_metrics.items()
				})
			else:
				env_metrics = self.benv.get_env_metrics(levels)
		else:
			env_metrics = None

		plr_stats = self.plr_mgr.get_metrics(train_state.plr_buffer)

		stats = self._compile_stats(update_stats, ep_stats, is_replay, env_metrics, plr_stats)

		if self.n_devices > 1:
			plr_buffer = train_state.plr_buffer
			train_state = train_state.replace(plr_buffer=None)

		train_state = train_state.increment()
		stats.update(dict(n_updates=train_state.n_updates[0]))

		return (
			stats, 
			rng, 
			train_state, 
			state, 
			start_state, 
			obs, 
			carry, 
			extra, 
			ep_stats,
			plr_buffer
		)
		


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import copy
from functools import partial
from collections import defaultdict
import time

import numpy as np
import jax
from jax.sharding import Mesh, PartitionSpec as P
from jax.experimental import mesh_utils
from jax.experimental.shard_map import shard_map

from .eval_runner import EvalRunner
from .dr_runner import DRRunner
from .paired_runner import PAIREDRunner
from .plr_runner import PLRRunner
from minimax.util.rl import UEDScore, PopPLRManager
import minimax.envs as envs
import minimax.models as models
import minimax.agents as agents


class RunnerInfo:
	def __init__(
		self, 
		runner_cls, 
		is_ued=False):
		self.runner_cls = runner_cls
		self.is_ued = is_ued


RUNNER_INFO = {
	'dr': RunnerInfo(
		runner_cls=DRRunner,
	),
	'plr': RunnerInfo(
		runner_cls=PLRRunner,
	),
	'paired': RunnerInfo(
		runner_cls=PAIREDRunner,
		is_ued=True
	)
}


class ExperimentRunner:
	def __init__(
		self,
		train_runner,
		env_name,
		agent_rl_algo,
		student_model_name,
		teacher_model_name=None,
		train_runner_kwargs={},
		env_kwargs={},
		ued_env_kwargs={},
		student_rl_kwargs={},
		teacher_rl_kwargs={},
		student_model_kwargs={},
		teacher_model_kwargs={},
		eval_kwargs={},
		eval_env_kwargs={},
		n_devices=1,
	):	
		self.env_name = env_name
		self.agent_rl_algo = agent_rl_algo
		self.is_ued = RUNNER_INFO[train_runner].is_ued

		dummy_env = envs.make(
			env_name, 
			env_kwargs, 
			ued_env_kwargs)[0]

		# ---- Make agent ---- 
		student_model_kwargs['output_dim'] = dummy_env.action_space().n
		student_model = models.make(
			env_name=env_name,
			model_name=student_model_name,
			**student_model_kwargs
		)

		student_agent = agents.PPOAgent(
			model=student_model,
			n_devices=n_devices,
			**student_rl_kwargs
		)

		# ---- Handle UED-related settings ---- 
		if self.is_ued:
			max_teacher_steps = dummy_env.ued_max_episode_steps()
			teacher_model_kwargs['n_scalar_embeddings'] = max_teacher_steps
			teacher_model_kwargs['max_scalar'] = max_teacher_steps
			teacher_model_kwargs['output_dim'] = dummy_env.ued_action_space().n

			teacher_model = models.make(
				env_name=env_name,
				model_name=teacher_model_name,
				**teacher_model_kwargs
			)

			teacher_agent = agents.PPOAgent(
				model=teacher_model,
				n_devices=n_devices,
				**teacher_rl_kwargs
			)

			train_runner_kwargs.update(dict(
				teacher_agents=[teacher_agent]
			))
			train_runner_kwargs.update(dict(
				ued_env_kwargs=ued_env_kwargs
			))

		# Debug, tabulate student and teacher model
		# import jax.numpy as jnp
		# dummy_rng = jax.random.PRNGKey(0)
		# obs, _ = dummy_env.reset(dummy_rng)
		# hx = student_model.initialize_carry(dummy_rng, (1,))
		# ued_obs, _ = dummy_env.reset_teacher(dummy_rng)
		# ued_hx = teacher_model.initialize_carry(dummy_rng, (1,))

		# obs['image'] = jnp.expand_dims(obs['image'], 0)
		# ued_obs['image'] = jnp.expand_dims(ued_obs['image'], 0)

		# print(student_model.tabulate(dummy_rng, obs, hx))
		# print(teacher_model.tabulate(dummy_rng, ued_obs, hx))

		# import pdb; pdb.set_trace()

		# ---- Set up train runner ----
		runner_cls = RUNNER_INFO[train_runner].runner_cls

		# Set up learning rate annealing parameters
		lr_init = train_runner_kwargs.lr
		lr_final = train_runner_kwargs.lr_final
		lr_anneal_steps = train_runner_kwargs.lr_anneal_steps

		if lr_final is None:
			train_runner_kwargs.lr_final = lr_init
		if train_runner_kwargs.lr_final == train_runner_kwargs.lr:
			train_runner_kwargs.lr_anneal_steps = 0

		self.runner = runner_cls(
			env_name=env_name,
			env_kwargs=env_kwargs,
			student_agents=[student_agent],
			n_devices=n_devices,
			**train_runner_kwargs)

		# ---- Make eval runner ----
		if eval_kwargs.get('env_names') is None:
			self.eval_runner = None
		else:
			self.eval_runner = EvalRunner(
				pop=self.runner.student_pop,
				env_kwargs=eval_env_kwargs,
				**eval_kwargs)

		self._start_tick = 0

		# ---- Set up device parallelism ----
		self.n_devices = n_devices
		if n_devices > 1:
			dummy_runner_state = self.reset_train_runner(jax.random.PRNGKey(0))
			self._shmap_run = self._make_shmap_run(dummy_runner_state)
		else:
			self._shmap_run = None

	@partial(jax.jit, static_argnums=(0,))
	def step(self, runner_state, evaluate=False):
		if self.n_devices > 1:
			run_fn = self._shmap_run
		else:
			run_fn = self.runner.run

		stats, *runner_state = run_fn(*runner_state)

		rng = runner_state[0]
		rng, subrng = jax.random.split(rng)

		if self.eval_runner is not None:
			params = runner_state[1].params
			eval_stats = jax.lax.cond(
				evaluate,
				self.eval_runner.run,
				self.eval_runner.fake_run,
				*(subrng, params)
			)
		else:
			eval_stats = {}

		return stats, eval_stats, rng, *runner_state[1:]

	def _make_shmap_run(self, runner_state):
		devices = mesh_utils.create_device_mesh((self.n_devices,))
		mesh = Mesh(devices, axis_names=('device'))

		in_specs, out_specs = self.runner.get_shmap_spec()

		return partial(shard_map,
			mesh=mesh,
			in_specs=in_specs,
			out_specs=out_specs,
			check_rep=False
		)(self.runner.run)

	def train(
		self, 
		rng,
		agent_algo='ppo',
		algo_runner='dr',
		n_total_updates=1000,
		logger=None,
		log_interval=1,
		test_interval=1,
		checkpoint_interval=0,
		archive_interval=0,
		archive_init_checkpoint=False,
		from_last_checkpoint=False
	):
		"""
		Entry-point for training
		"""
		# Load checkpoint if any
		runner_state = self.runner.reset(rng)

		if from_last_checkpoint:
			last_checkpoint_state = logger.load_last_checkpoint_state()
			if last_checkpoint_state is not None:
				runner_state = self.runner.load_checkpoint_state(
					runner_state, 
					last_checkpoint_state
				)
				self._start_tick = runner_state[1].n_iters[0]

		# Archive initialization weights if necessary
		if archive_init_checkpoint:
			logger.checkpoint(
				self.runner.get_checkpoint_state(runner_state), 
				index=0, 
				archive_interval=1)

		# Train loop
		log_on = logger is not None and log_interval > 0
		checkpoint_on = checkpoint_interval > 0 or archive_interval > 0
		train_state = runner_state[1]

		tick = self._start_tick
		train_steps = tick*self.runner.step_batch_size*self.runner.n_rollout_steps*self.n_devices
		real_train_steps = train_steps//self.runner.n_students

		while (train_state.n_updates < n_total_updates).any():
			evaluate = test_interval > 0 and (tick+1) % test_interval == 0

			start = time.time()
			stats, eval_stats, *runner_state = \
				self.step(runner_state, evaluate)
			end = time.time()

			if evaluate:
				stats.update(eval_stats)
			else:
				stats.update({k:None for k in eval_stats.keys()})

			train_state = runner_state[1]
			
			dsteps = self.runner.step_batch_size*self.runner.n_rollout_steps*self.n_devices
			real_dsteps = dsteps//self.runner.n_students
			train_steps += dsteps
			real_train_steps += real_dsteps
			sps = int(dsteps/(end-start))
			real_sps = int(real_dsteps/(end-start))
			stats.update(dict(
				steps=train_steps,
				sps=sps,
				real_steps=real_train_steps,
				real_sps=real_sps
			))

			tick += 1

			if log_on and tick % log_interval == 0:
				logger.log(stats, tick, ignore_val=-np.inf)

			if checkpoint_on and tick > 0:
				if tick % checkpoint_interval == 0 \
					or (archive_interval > 0 and tick % archive_interval == 0):
					checkpoint_state = \
						self.runner.get_checkpoint_state(runner_state)
					logger.checkpoint(
						checkpoint_state, 
						index=tick, 
						archive_interval=archive_interval)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from abc import ABC


class Agent(ABC):
    """
    Generic interface for an agent.
    """
    @property
    def is_recurrent(self):
        pass

    @property
    def action_info_keys(self):
    	pass

    def init_params(self, rng, obs, carry=None):
        pass

    def init_carry(self, rng, batch_dims):
        pass 

    def act(self, *args, **kwargs):
        pass

    def get_action_dist(self, dist_params, dtype):
        pass

    def evaluate(self, *args, **kwargs):
        pass

    def update(self, *args, **kwargs):
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import jax
import chex
from typing import Tuple, Union, Optional
from functools import partial
from flax import struct

from .environment import EnvParams, EnvState


class UEDEnvironment:
	"""
	Wraps two Environment instances, one being the basic environment,
	and the other, its UED counterpart.

	The interface extends the student environment interace.
	"""
	def __init__(self, env, ued_env):
		self.env = env
		self.ued_env = ued_env

		# Default reset and step centers on student
		self.reset = self.reset_random 
		self.step = self.env.step

	def reset_random(
		self, 
		rng: chex.PRNGKey, 
	) -> Tuple[chex.ArrayTree, EnvState]:
		return self.env.reset(rng)

	def reset_teacher(
		self, 
		rng: chex.PRNGKey, 
	) -> Tuple[chex.ArrayTree, EnvState]:
		return self.ued_env.reset(rng)

	def step_teacher(
		self,
		rng: chex.PRNGKey,
		ued_state: EnvState,
		action: Union[int, float],
	) -> Tuple[chex.ArrayTree, EnvState, float, bool, dict]:
		return self.ued_env.step(
			rng, ued_state, action, reset_on_done=False)

	def reset_student(
		self,
		rng: chex.PRNGKey, 
		ued_state: EnvState,
	) -> Tuple[chex.ArrayTree, EnvState]:
		"""
		Reset the student based on 
		"""
		encoding = self.ued_env.get_env_instance(rng, ued_state)
		return self.env.set_env_instance(encoding)

	def step_student(
		self,
		rng: chex.PRNGKey,
		state: EnvState,
		action: Union[int, float],
		reset_state: Optional[chex.ArrayTree] = None
	) -> Tuple[chex.ArrayTree, EnvState, float, bool, dict]:
		return self.env.step(
			rng, 
			state, 
			action, 
			reset_state=reset_state)

	def set_env_instance(self, encoding: chex.ArrayTree):
		return self.env.set_env_instance(encoding) 

	# Spaces interface
	def action_space(self):
		"""Action space of the environment."""
		return self.env.action_space()

	def observation_space(self):
		"""Observation space of the environment."""
		return self.env.observation_space()

	def state_space(self):
		"""Observation space of the environment."""
		return self.env.state_space()

	def max_episode_steps(self):
		"""Action space of the environment."""
		return self.env.max_episode_steps()

	def ued_action_space(self):
		"""Action space of the environment."""
		return self.ued_env.action_space()

	def ued_observation_space(self):
		"""Observation space of the environment."""
		return self.ued_env.observation_space()

	def ued_state_space(self):
		"""Observation space of the environment."""
		return self.ued_env.state_space()

	def ued_max_episode_steps(self):
		"""Action space of the environment."""
		return self.ued_env.max_episode_steps()

	def get_env_metrics(self, state: EnvState):
		"""Environment-specific metrics, e.g. number of walls."""
		return self.env.get_env_metrics(state)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import os
import shutil
import pickle
import json
from pathlib import Path

from .dotdict import DefaultDotDict


def save_pkl_object(obj, path):
    """Helper to store pickle objects."""
    output_file = Path(path)
    output_file.parent.mkdir(exist_ok=True, parents=True)

    with open(path, "wb") as output:
        # Overwrites any existing file.
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)

    print(f"Stored checkpoint at {path}.")


def load_pkl_object(path: str):
    """Helper to reload pickle objects."""
    with open(path, "rb") as input:
        obj = pickle.load(input)
    print(f"Loaded checkpoint from {path}.")
    return obj


def safe_checkpoint(
    state, 
    dir_path, 
    name, 
    index=None, 
    archive_interval=None):
    savename = f'{name}.pkl'
    tmp_savepath = f'{name}_tmp.pkl'

    save_path = os.path.join(dir_path, savename)
    tmp_savepath = os.path.join(dir_path, tmp_savepath)

    save_pkl_object(state, tmp_savepath)

    # Rename
    os.replace(tmp_savepath, save_path)

    # Archive if needed
    if index is not None and archive_interval is not None and archive_interval > 0:
        if index % archive_interval == 0:
            archive_path = os.path.join(dir_path, f'{name}_{index}.pkl')
            shutil.copy(save_path, archive_path)


def load_config(path: str):
    with open(path) as meta_file:
        _config = json.load(meta_file)['config']

    config = {}
    for k,v in _config.items():
        if isinstance(v, dict):
            v = DefaultDotDict(v)
        config[k] = v

    return DefaultDotDict(config)





"""
Copyright 2018 The JAX Authors.

This file is based on the OptimizedLSTMCell class from
https://github.com/google/jax

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

	https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from functools import partial
from typing import Any, Tuple, Mapping

import numpy as np
import jax.numpy as jnp
import flax.linen as nn
from flax.linen.dtypes import promote_dtype
from flax.linen.module import compact
from flax.linen.recurrent import DenseParams


Array = Any


class CustomOptimizedLSTMCell(nn.OptimizedLSTMCell):
	@compact
	def __call__(self, carry: Tuple[Array, Array],
		inputs: Array) -> Tuple[Tuple[Array, Array], Array]:
		r"""An optimized long short-term memory (LSTM) cell.

		Args:
		  carry: the hidden state of the LSTM cell, initialized using
			`LSTMCell.initialize_carry`.
		  inputs: an ndarray with the input for the current time step. All
			dimensions except the final are considered batch dimensions.

		Returns:
		  A tuple with the new carry and the output.
		"""
		c, h = carry
		hidden_features = h.shape[-1]

		def _concat_dense(inputs: Array,
						  params: Mapping[str, Tuple[Array, Array]],
						  use_bias: bool = True) -> Array:
			# Concatenates the individual kernels and biases, given in params, into a
			# single kernel and single bias for efficiency before applying them using
			# dot_general.
			kernels, biases = zip(*params.values())
			kernel = jnp.concatenate(kernels, axis=-1)
			if use_bias:
				bias = jnp.concatenate(biases, axis=-1)
			else:
				bias = None
			inputs, kernel, bias = promote_dtype(inputs, kernel, bias, dtype=self.dtype)
			y = jnp.dot(inputs, kernel)
			if use_bias:
				y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))

			# Split the result back into individual (i, f, g, o) outputs.
			split_indices = np.cumsum([kernel.shape[-1] for kernel in kernels[:-1]])
			ys = jnp.split(y, split_indices, axis=-1)
			return dict(zip(params.keys(), ys))

		# Create params with the same names/shapes as `LSTMCell` for compatibility.
		dense_params_h = {}
		dense_params_i = {}
		for component in ['i', 'f', 'g', 'o']:
			dense_params_i[component] = DenseParams(
				features=hidden_features, use_bias=True,
				param_dtype=self.param_dtype,
				kernel_init=self.kernel_init, bias_init=self.bias_init,
				name=f'i{component}')(inputs)
			dense_params_h[component] = DenseParams(
				features=hidden_features, use_bias=True,
				param_dtype=self.param_dtype,
				kernel_init=self.recurrent_kernel_init, bias_init=self.bias_init,
				name=f'h{component}')(h)
		dense_h = _concat_dense(h, dense_params_h, use_bias=True)
		dense_i = _concat_dense(inputs, dense_params_i, use_bias=True)

		i = self.gate_fn(dense_h['i'] + dense_i['i'])
		f = self.gate_fn(dense_h['f'] + dense_i['f'])
		g = self.activation_fn(dense_h['g'] + dense_i['g'])
		o = self.gate_fn(dense_h['o'] + dense_i['o'])

		new_c = f * c + i * g
		new_h = o * self.activation_fn(new_c)
		return (new_c, new_h), new_h



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from typing import Tuple

import numpy as np
import jax
import jax.numpy as jnp
import flax.linen as nn
import chex
from tensorflow_probability.substrates import jax as tfp

from minimax.models import common
from minimax.models import s5
from minimax.models.registration import register


class GridWorldBasicModel(nn.Module):
	"""Split Actor-Critic Architecture for PPO."""
	output_dim: int = 7
	n_hidden_layers: int = 1
	hidden_dim: int = 32
	n_conv_filters: int = 16
	conv_kernel_size: int = 3
	n_scalar_embeddings: int = 4
	max_scalar: int = 4
	scalar_embed_dim: int = 5
	recurrent_arch: str = None
	recurrent_hidden_dim: int = 256
	base_activation: str = 'relu'
	head_activation: str = 'tanh'

	s5_n_blocks: int = 2
	s5_n_layers: int = 4
	s5_layernorm_pos: str = None
	s5_activation: str = "half_glu1"

	value_ensemble_size: int = 1

	def setup(self):
		self.conv = nn.Sequential([
			nn.Conv(
				features=self.n_conv_filters, 
				kernel_size=[self.conv_kernel_size,]*2,
				strides=1, 
				kernel_init=common.init_orth(
					scale=common.calc_gain(self.base_activation)
				),
				padding='VALID',
				name='cnn'),
			common.get_activation(self.base_activation)
		])

		if self.n_scalar_embeddings > 0:
			self.fc_scalar = nn.Embed(
				num_embeddings=self.n_scalar_embeddings,
				features=self.scalar_embed_dim, 
				embedding_init=common.init_orth(
					common.calc_gain('linear')
				),
				name=f'fc_scalar'
			)
		elif self.scalar_embed_dim > 0:
			self.fc_scalar = nn.Dense(
				self.scalar_embed_dim,
				kernel_init=common.init_orth(
					common.calc_gain('linear')
				),
				name=f'fc_scalar'
			)
		else:
			self.fc_scalar = None

		if self.recurrent_arch is not None:
			if self.recurrent_arch == 's5':
				self.embed_pre_s5 = nn.Sequential([
					nn.Dense(
						self.recurrent_hidden_dim,
						kernel_init=common.init_orth(
							common.calc_gain('linear')
						),
						name=f'fc_pre_s5'
						)
				])
				self.rnn = s5.make_s5_encoder_stack(
					input_dim=self.recurrent_hidden_dim,
					ssm_state_dim=self.recurrent_hidden_dim,
					n_blocks=self.s5_n_blocks,
					n_layers=self.s5_n_layers,
					activation=self.s5_activation,
					layernorm_pos=self.s5_layernorm_pos
				)
			else:
				self.rnn = common.ScannedRNN(
					recurrent_arch=self.recurrent_arch,
					recurrent_hidden_dim=self.recurrent_hidden_dim,
					kernel_init=common.init_orth(),
					recurrent_kernel_init=common.init_orth()
				)
		else:
			self.rnn = None

		self.pi_head = nn.Sequential([
			common.make_fc_layers(
				'fc_pi', 
				n_layers=self.n_hidden_layers,
				hidden_dim=self.hidden_dim,
				activation=common.get_activation(self.head_activation),
				kernel_init=common.init_orth(
					common.calc_gain(self.head_activation)
				)
			),
			nn.Dense(
				self.output_dim, 
				kernel_init=nn.initializers.constant(0.01), 
				name=f'fc_pi_final'
			)
		])

		value_head_kwargs = dict(
			n_hidden_layers=self.n_hidden_layers, 
			hidden_dim=self.hidden_dim,
			activation=nn.tanh,
			kernel_init=common.init_orth(
				common.calc_gain('tanh')
			),
			last_layer_kernel_init=common.init_orth(
				common.calc_gain('linear')
			)
		)

		if self.value_ensemble_size > 1:
			self.v_head = common.EnsembleValueHead(
				n=self.value_ensemble_size, **value_head_kwargs)
		else:
			self.v_head = common.ValueHead(**value_head_kwargs)

	def __call__(self, x, carry=None):
		raise NotImplementedError

	def initialize_carry(
			self, 
			rng: chex.PRNGKey, 
			batch_dims: Tuple[int] = ()) -> Tuple[chex.ArrayTree, chex.ArrayTree]:
		"""Initialize hidden state of LSTM."""
		if self.recurrent_arch is not None:
			if self.recurrent_arch == 's5':
				return s5.S5EncoderStack.initialize_carry( # Since conj_sym=True
					rng, batch_dims, self.recurrent_hidden_dim//2, self.s5_n_layers
				)
			else:
				return common.ScannedRNN.initialize_carry(
					rng, batch_dims, self.recurrent_hidden_dim, self.recurrent_arch)
		else:
			raise ValueError('Model is not recurrent.')

	@property
	def is_recurrent(self):
		return self.recurrent_arch is not None


class GridWorldACStudentModel(GridWorldBasicModel):
	def __call__(self, x, carry=None, reset=None):
		"""
		Inputs:
			x: B x h x w observations
			hxs: B x hx_dim hidden states
			masks: B length vector of done masks
		"""
		old_x = x
		img = x['image']
		agent_dir = x['agent_dir']
		aux = x.get('aux')

		if self.rnn is not None:
			batch_dims = img.shape[:2]
			x = self.conv(img).reshape(*batch_dims, -1)
		else:
			batch_dims = img.shape[:1]
			x = self.conv(img).reshape(*batch_dims, -1)

		if self.fc_scalar is not None:
			if self.n_scalar_embeddings == 0:
				agent_dir /= self.max_scalar

			scalar_emb = self.fc_scalar(agent_dir).reshape(*batch_dims, -1)
			x = jnp.concatenate([x, scalar_emb], axis=-1)

		if aux is not None:
			x = jnp.concatenate([x, aux], axis=-1)
			
		if self.rnn is not None:
			if self.recurrent_arch == 's5':
				x = self.embed_pre_s5(x)
				carry, x = self.rnn(carry, x, reset)
			else:
				carry, x = self.rnn(carry, (x, reset))

		logits = self.pi_head(x)

		v = self.v_head(x)

		return v, logits, carry


class GridWorldACTeacherModel(GridWorldBasicModel):
	"""
	Original teacher model from Dennis et al, 2020. It is identical ins
	high-level spec to the student model, but with the additional fwd logic
	of concatenating a noise vector.
	"""
	def __call__(self, x, carry=None, reset=None):
		"""
		Inputs:
			x: B x h x w observations
			hxs: B x hx_dim hidden states
			masks: B length vector of done masks
		"""
		img = x['image']
		time = x['time']
		noise = x.get('noise')
		aux = x.get('aux')

		if self.rnn is not None:
			batch_dims = img.shape[:2]
			x = self.conv(img).reshape(*batch_dims, -1)
		else:
			batch_dims = img.shape[:1]
			x = self.conv(img).reshape(*batch_dims, -1)

		if self.fc_scalar is not None:
			if self.n_scalar_embeddings == 0:
				time /= self.max_scalar

			scalar_emb = self.fc_scalar(time).reshape(*batch_dims, -1)
			x = jnp.concatenate([x, scalar_emb], axis=-1)

		if noise is not None:
			noise = noise.reshape(*batch_dims, -1)
			x = jnp.concatenate([x, noise], axis=-1)

		if aux is not None:
			x = jnp.concatenate([x, aux], axis=-1)

		if self.rnn is not None:
			if self.recurrent_arch == 's5':
				x = self.embed_pre_s5(x)
				carry, x = self.rnn(carry, x, reset)
			else:
				carry, x = self.rnn(carry, (x, reset))

		logits = self.pi_head(x)

		v = self.v_head(x)

		return v, logits, carry


# Register models
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register(
	env_group_id='Maze', model_id='default_student_cnn', 
	entry_point=module_path + ':GridWorldACStudentModel')

register(
	env_group_id='Maze', model_id='default_teacher_cnn', 
	entry_point=module_path + ':GridWorldACTeacherModel')



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file is modified from 
https://github.com/Farama-Foundation/Minigrid

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0
"""

import sys
import numpy as np

# Only ask users to install matplotlib if they actually need it
try:
    import matplotlib.pyplot as plt
except:
    print('To display the environment in a window, please install matplotlib, eg:')
    print('pip3 install --user matplotlib')
    sys.exit(-1)

class Window:
    """
    Window to draw a gridworld instance using Matplotlib
    """

    def __init__(self, title):
        self.fig = None

        self.imshow_obj = None

        # Create the figure and axes
        self.fig, self.ax = plt.subplots()

        # Show the env name in the window title
        self.fig.canvas.manager.set_window_title(title)

        # Turn off x/y axis numbering/ticks
        self.ax.set_xticks([], [])
        self.ax.set_yticks([], [])

        # Flag indicating the window was closed
        self.closed = False

        def close_handler(evt):
            self.closed = True

        self.fig.canvas.mpl_connect('close_event', close_handler)

    def show_img(self, img):
        """
        Show an image or update the image being shown
        """

        # Show the first image of the environment
        if self.imshow_obj is None:
            self.imshow_obj = self.ax.imshow(img, interpolation='bilinear')

        self.imshow_obj.set_data(img)
        self.fig.canvas.draw()

        # Let matplotlib process UI events
        # This is needed for interactive mode to work properly
        plt.pause(0.001)

    def save_img(self, path):
    	plt.savefig(path, bbox_inches='tight', pad_inches=0)

    def set_caption(self, text):
        """
        Set/update the caption text below the image
        """

        plt.xlabel(text)

    def reg_key_handler(self, key_handler):
        """
        Register a keyboard event handler
        """

        # Keyboard handler
        self.fig.canvas.mpl_connect('key_press_event', key_handler)

    def show(self, block=True):
        """
        Show the window, and start an event loop
        """

        # If not blocking, trigger interactive mode
        if not block:
            plt.ion()

        # Show the plot
        # In non-interative mode, this enters the matplotlib event loop
        # In interactive mode, this call does not block
        plt.show()

    def close(self):
        """
        Close the window
        """

        plt.close()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse

from minimax.util.parsnip import Parsnip
from minimax.util.args import str2bool


parser = Parsnip()


# ==== Define top-level arguments
parser.add_argument(
    '--seed', 
    type=int, 
    default=1, 
    help='Training seed.')
parser.add_argument(
    '--agent_rl_algo',
    type=str,
    default='ppo',
    choices=['ppo'],
    help='Base RL algorithm to use.')
parser.add_argument(
    '--n_total_updates',
    type=int,
    default=30000,
    help='Total number of student gradient updates.')
parser.add_argument(
    '--train_runner',
    type=str,
    default='dr',
    choices=['dr', 'plr', 'paired'],
    help='Algorithm runner.')
parser.add_argument(
    '--n_devices',
    type=int,
    default=1,
    help='Number of devices.')


# ==== RL runner arguments.
train_runner_subparser = parser.add_subparser(
    name='train_runner')
train_runner_subparser.add_argument(
    '--n_students', 
    type=int, 
    default=1, 
    help='Number of students in population.')
train_runner_subparser.add_argument(
    '--n_parallel', 
    type=int, 
    default=1, 
    help='Number of parallel environments per rollout.')
train_runner_subparser.add_argument(
    '--n_eval', 
    type=int, 
    default=1, 
    help='Number of student evaluations per environment.')
train_runner_subparser.add_argument(
    '--n_rollout_steps', 
    type=int, 
    default=250, 
    help='Number of rollout steps.')
train_runner_subparser.add_argument(
    '--lr', 
    type=float, 
    default=1e-4, 
    help='Initial learning rate.')
train_runner_subparser.add_argument(
    '--lr_final', 
    type=float, 
    default=None,
    nargs="?", 
    help='Final learning rate.')
train_runner_subparser.add_argument(
    '--lr_anneal_steps', 
    type=int, 
    default=0,
    nargs="?", 
    help='Number of learning rate annealing steps.')
train_runner_subparser.add_argument(
    '--max_grad_norm',
    type=float,
    default=0.5,
    help='max norm of gradients.')
train_runner_subparser.add_argument(
    '--adam_eps',
    type=float,
    default=1e-5,
    help='Adam eps.')
train_runner_subparser.add_argument(
    '--track_env_metrics', 
    type=str2bool,
    default=False,
    help='Track env metrics during training. Can reduce SPS.')
train_runner_subparser.add_argument(
    '--discount',
    type=float,
    default=0.995,
    help='Student discount factor for rewards')
train_runner_subparser.add_argument(
    '--n_unroll_rollout',
    type=int,
    default=1,
    help='Number of times to unroll rollout scan.')
train_runner_subparser.add_argument(
    '--render',
    type=str2bool,
    default=False,
    help='Whether to render.')

# ------ AC-specific arguments -----
dr_subparser = parser.add_subparser(
    name='dr',
    prefix='dr',
    dependency={'train_runner': 'dr'},
    dest='train_runner')

# -------- General UED arguments --------
parser.add_dependent_argument(
    '--ued_score',
    type=str,
    default='relative_regret',
    dependency={'train_runner': ['plr', 'paired']},
    dest='train_runner',
    choices=[
        'relative_regret',
        'mean_relative_regret',
        'population_regret', 
        'neg_return', # aka minimax adversarial
        'l1_value_loss', 
        'positive_value_loss',
        'max_mc',
        'value_disagreement'
    ],
    help='UED score of agent.')

# -------- PAIRED arguments --------
plr_subparser = parser.add_subparser(
    name='plr',
    prefix='plr',
    dependency={'train_runner': 'plr'},
    dest='train_runner')
plr_subparser.add_argument(
    '--replay_prob',
    type=float,
    default=0.5,
    help='PLR replay probability.'
)
plr_subparser.add_argument(
    '--buffer_size',
    type=int,
    default=128,
    help='PLR level buffer size.'
)
plr_subparser.add_argument(
    '--staleness_coef',
    type=float,
    default=0.3,
    help='Staleness coefficient.'
)
plr_subparser.add_argument(
    '--temp',
    type=float,
    default=1.0,
    help='Score distribution temperature.'
)
plr_subparser.add_argument(
    '--use_score_ranks',
    type=str2bool,
    default=True,
    help='Use rank-based prioritiziation.'
)
plr_subparser.add_argument(
    '--min_fill_ratio',
    type=float,
    default=0.5,
    help='Minimum fill ratio before level replay begins.'
)
plr_subparser.add_argument(
    '--use_robust_plr',
    type=str2bool,
    default=True,
    help='Use robust PLR.'
)
plr_subparser.add_argument(
    '--use_parallel_eval',
    type=str2bool,
    default=False,
    help='Use rank-based prioritiziation.'
)
plr_subparser.add_argument(
    '--force_unique',
    type=str2bool,
    default=False,
    help='Force level buffer members to be unique.'
)
plr_subparser.add_argument(
    '--mutation_fn',
    type=str,
    default=None,
    help='Name of mutation function for ACCEL.'
)
plr_subparser.add_argument(
    '--n_mutations',
    type=int,
    default=0,
    help='Number of mutations per iteration of ACCEL.'
)
plr_subparser.add_argument(
    '--mutation_criterion',
    type=str,
    default='batch',
    help='Criterion for choosing PLR buffer members to mutate.'
)
plr_subparser.add_argument(
    '--mutation_subsample_size',
    type=int,
    default=0,
    help='Number of PLR buffer members to mutate into a full batch.'
)


# -------- PAIRED arguments --------
paired_subparser = parser.add_subparser(
    name='paired',
    prefix='paired',
    dependency={'train_runner': 'paired'},
    dest='train_runner')


# ==== Student RL arguments.
student_rl_subparser = parser.add_subparser(
    name='student_rl',
    prefix='student')
student_rl_subparser.add_argument(
    '--entropy_coef',
    type=float,
    default=0.0,
    help='entropy term coefficient')
student_rl_subparser.add_argument(
    '--value_loss_coef',
    type=float,
    default=0.5,
    help='value loss coefficient (default: 0.5)')
student_rl_subparser.add_argument(
    '--n_unroll_update',
    type=int,
    default=1,
    help='Number of times to unroll minibatch scan.')

# -------- Student PPO arguments. --------
student_ppo_subparser = parser.add_subparser(
    name='student_ppo',
    prefix='student_ppo',
    dest='student_rl',
    dependency={'agent_rl_algo': 'ppo'})
student_ppo_subparser.add_argument(
    '--n_epochs',
    type=int,
    default=5,
    help='Number of PPO epochs.')
student_ppo_subparser.add_argument(
    '--n_minibatches',
    type=int,
    default=1,
    help='Number of minibatches per PPO epoch.')
student_ppo_subparser.add_argument(
    '--clip_eps',
    type=float,
    default=0.2,
    help='PPO clip parameter')
student_ppo_subparser.add_argument(
    '--clip_value_loss',
    type=str2bool,
    default=True,
    help='ppo clip value loss')
parser.add_dependent_argument(
    '--gae_lambda',
    type=float,
    default=0.95,
    prefix='student',
    dependency={'agent_rl_algo': 'ppo'},
    dest='train_runner',
    help='GAE lambda parameter for student.')


# ==== Teacher RL arguments.
teacher_rl_subparser = parser.add_subparser(
    name='teacher_rl',
    prefix='teacher',
    dependency={'train_runner':['paired']})
teacher_rl_subparser.add_argument(
    '--entropy_coef',
    type=float,
    default=0.0,
    help='entropy term coefficient')
teacher_rl_subparser.add_argument(
    '--value_loss_coef',
    type=float,
    default=0.5,
    help='value loss coefficient (default: 0.5)')
teacher_rl_subparser.add_argument(
    '--n_unroll_update',
    type=int,
    default=1,
    help='Number of times to unroll minibatch scan.')
parser.add_dependent_argument(
    '--teacher_discount',
    type=float,
    default=0.995,
    dependency={'train_runner': 'paired'},
    dest='train_runner',
    help='discount factor for rewards')
parser.add_dependent_argument(
    '--teacher_lr',
    type=float, 
    default=None,
    nargs="?",
    dependency={'agent_rl_algo': 'ppo', 'train_runner': 'paired'},
    dest='train_runner',
    help='Initial learning rate of teacher.')
parser.add_dependent_argument(
    '--teacher_lr_final',
    type=float,
    default=None,
    nargs="?",
    dependency={'agent_rl_algo': 'ppo', 'train_runner': 'paired'},
    dest='train_runner',
    help='Initial learning rate of teacher.')
parser.add_dependent_argument(
    '--teacher_lr_anneal_steps',
    type=int,
    default=0,
    nargs="?",
    dependency={'agent_rl_algo': 'ppo', 'train_runner': 'paired'},
    dest='train_runner',
    help='Initial learning rate of teacher.')


# -------- Teacher PPO arguments. --------
teacher_ppo_subparser = parser.add_subparser(
    name='teacher_ppo',
    prefix='teacher_ppo',
    dest='teacher_rl',
    dependency={'agent_rl_algo': 'ppo', 'train_runner': 'paired'})
teacher_ppo_subparser.add_argument(
    '--n_epochs',
    type=int,
    default=5,
    help='Number of PPO epochs.')
teacher_ppo_subparser.add_argument(
    '--n_minibatches',
    type=int,
    default=1,
    help='Number of minibatches per PPO epoch.')
teacher_ppo_subparser.add_argument(
    '--clip_eps',
    type=float,
    default=0.2,
    help='PPO clip parameter')
teacher_ppo_subparser.add_argument(
    '--clip_value_loss',
    type=str2bool,
    default=True,
    help='ppo clip value loss')
parser.add_dependent_argument(
    '--teacher_gae_lambda',
    type=float,
    default=0.95,
    dependency={'agent_rl_algo': 'ppo', 'train_runner': 'paired'},
    dest='train_runner',
    help='GAE lambda parameter for teacher.')


# ==== Student model arguments.
parser.add_argument(
    '--student_model_name',
    type=str,
    default='default_student_cnn',
    help='Name of student model architecture.')

# Placeholder group for student model args
student_model_parser = parser.add_subparser(
    name='student_model',
    prefix='student')

# ---- Maze args for student model ----
student_maze_model_parser = parser.add_subparser(
    name='student_maze_model',
    prefix='student',
    dest="student_model",
    dependency={'env_name': 'Maze*'})
student_maze_model_parser.add_argument(
    '--recurrent_arch',
    type=str,
    default=None,
    nargs='?',
    choices=['gru', 'lstm', 's5'],
    help='Student RNN architecture.')
student_maze_model_parser.add_argument(
    '--recurrent_hidden_dim',
    type=int,
    default=0,
    help='Student recurrent hidden state size.')
student_maze_model_parser.add_argument(
    '--hidden_dim',
    type=int,
    default=32,
    help='Student hidden dimension.')
student_maze_model_parser.add_argument(
    '--n_hidden_layers',
    type=int,
    default=1,
    help='Student number of hidden layers in policy/value heads.')
student_maze_model_parser.add_argument(
    '--n_conv_filters',
    type=int,
    default=16,
    help='Number of CNN filters for student.')
student_maze_model_parser.add_argument(
    '--n_scalar_embeddings',
    type=int,
    default=4,
    help='Defaults to 4 directional embeddings.')
student_maze_model_parser.add_argument(
    '--scalar_embed_dim',
    type=int,
    default=5,
    help='Dimensionality of scalar direction embeddings.')
student_maze_model_parser.add_argument(
    '--base_activation',
    type=str,
    default='relu',
    choices=['relu', 'gelu', 'crelu', 'leaky_relu'],
    help='Nonlinearity for intermediate layers.')
student_maze_model_parser.add_argument(
    '--value_ensemble_size',
    type=int,
    default=1,
    help='Size of value ensemble. Defaults to 1 (no ensemble).')
student_maze_model_parser.add_argument(
    '--s5_n_blocks',
    type=int,
    default=1,
    help='Number of S5 blocks.')
student_maze_model_parser.add_argument(
    '--s5_n_layers',
    type=int,
    default=4,
    help='Number of S5 encoder layers.')
student_maze_model_parser.add_argument(
    '--s5_layernorm_pos',
    type=str,
    default=None,
    help='Layernorm pos in S5.')
student_maze_model_parser.add_argument(
    '--s5_activation',
    type=str,
    default="half_glu1",
    choices=["half_glu1", "half_glu2", "full_glu", "gelu"],
    help='Number of S5 encoder layers.')


# ==== Teacher model arguments.
parser.add_dependent_argument(
    '--teacher_model_name',
    dependency={'train_runner': ['paired']},
    type=str,
    help='Name of teacher model architecture.'
)

# Placeholder group for teacher model args
teacher_model_parser = parser.add_subparser(
    name='teacher_model',
    prefix='teacher',
    dependency={'train_runner': ['paired']})

# ---- Maze args for PAIRED teacher model ----
teacher_maze_model_parser = parser.add_subparser(
    name='teacher_maze_model',
    prefix='teacher',
    dest="teacher_model",
    dependency={'train_runner': 'paired', 'env_name': 'Maze*'})
teacher_maze_model_parser.add_argument(
    '--recurrent_arch',
    type=str,
    default=None,
    nargs='?',
    choices=['gru', 'lstm', 's5'],
    help='Teacher RNN architecture.')
teacher_maze_model_parser.add_argument(
    '--recurrent_hidden_dim',
    type=int,
    default=0,
    help='Teacher recurrent hidden state size.')
teacher_maze_model_parser.add_argument(
    '--hidden_dim',
    type=int,
    default=32,
    help='Teacher hidden dimension.')
teacher_maze_model_parser.add_argument(
    '--n_hidden_layers',
    type=int,
    default=1,
    help='Teacher number of hidden layers in policy/value heads.')
teacher_maze_model_parser.add_argument(
    '--n_conv_filters',
    type=int,
    default=128,
    help='Number of CNN filters for teacher.')
teacher_maze_model_parser.add_argument(
    '--scalar_embed_dim',
    type=int,
    default=10,
    help='Dimensionality of time-step embeddings.')
teacher_maze_model_parser.add_argument(
    '--base_activation',
    type=str,
    default='relu',
    choices=['relu', 'gelu', 'crelu', 'leaky_relu'],
    help='Nonlinearity for intermediate layers.')
teacher_maze_model_parser.add_argument(
    '--s5_n_blocks',
    type=int,
    default=1,
    help='Number of S5 blocks.')
teacher_maze_model_parser.add_argument(
    '--s5_n_layers',
    type=int,
    default=4,
    help='Number of S5 encoder layers.')
teacher_maze_model_parser.add_argument(
    '--s5_layernorm_pos',
    type=str,
    default=None,
    help='Layernorm pos in S5.')
teacher_maze_model_parser.add_argument(
    '--s5_activation',
    type=str,
    default="half_glu1",
    choices=["half_glu1", "half_glu2", "full_glu", "gelu"],
    help='Number of S5 encoder layers.')


# ==== Environment arguments.
parser.add_argument(
    '--env_name',
    type=str,
    default='Maze',
    help='Environment to train on')
env_parser = parser.add_subparser(
    name='env')

# -------- UED environment arguments. --------
ued_env_parser = parser.add_subparser(
    name='ued_env')

# ======== Envoronment-specific subparsers ======== 
# -------- Maze --------
env_maze_parser = parser.add_subparser(
    name='maze',
    prefix='maze',
    dependency={'env_name': ['Maze', 'Maze-MemoryMaze']},
    dest='env')
env_maze_parser.add_argument(
    '--height',
    type=int,
    default=13,
    help='Height of training mazes.')
env_maze_parser.add_argument(
    '--width',
    type=int,
    default=13,
    help='Width of training mazes.')
env_maze_parser.add_argument(
    '--n_walls',
    type=int,
    default=25,
    help='Maximum number of walls in training mazes.')
env_maze_parser.add_argument(
    '--replace_wall_pos',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Sample wall positions with replacement.')
env_maze_parser.add_argument(
    '--sample_n_walls',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Uniformly sample n_walls between 0 and n_walls.')
# -------- Maze* environments --------
env_maze_all_parser = parser.add_subparser(
    name='maze_all',
    prefix='maze',
    dependency={'env_name': 'Maze*'},
    dest='env')
env_maze_all_parser.add_argument(
    '--see_agent',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=True,
    help='Whether the agent sees itself in observations.')
env_maze_all_parser.add_argument(
    '--normalize_obs',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=True,
    help='Ensure observations are between 0 and 1.')
env_maze_all_parser.add_argument(
    '--obs_agent_pos',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Include agent xy pos in observations.')
env_maze_all_parser.add_argument(
    '--max_episode_steps',
    type=int,
    default=250,
    help='Maximum number of steps in training episodes.')

# -------- Maze UED --------
maze_ued_parser = parser.add_subparser(
    name='maze_ued',
    prefix='maze_ued',
    dependency={'env_name': ['Maze', 'Maze-MemoryMaze'], 'train_runner': 'paired'},
    dest='ued_env')
maze_ued_parser.add_argument(
    '--replace_wall_pos',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Teacher can sample same wall pos multiple times (resulting in variable n_walls).')
maze_ued_parser.add_argument(
    '--fixed_n_wall_steps',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Teacher samples exactly n_walls wall positions for each level.')
maze_ued_parser.add_argument(
    '--first_wall_pos_sets_budget',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='The first wall positional index determines the wall budget.')
maze_ued_parser.add_argument(
    '--noise_dim',
    type=int,  
    default=50,
    help="Dimension of episodic noise vector injected into the teacher's observation.")
maze_ued_parser.add_argument(
    '--n_walls',
    type=int,  
    default=25,
    help="Number walls the adversary can place.")
maze_ued_parser.add_argument(
    '--set_agent_dir',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help='Teacher chooses the agent direction on last time step.')
maze_ued_parser.add_argument(
    '--normalize_obs',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=True,
    help='Normalize teacher observations.')


# Logging arguments (All top-level arguments.).
parser.add_argument(
    "--verbose", 
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help="Print progress to stdout.")
parser.add_argument(
    '--xpid',
    default='latest',
    help='name for the run - prefix to log files')
parser.add_argument(
    '--log_dir',
    default='~/logs/minimax/',
    help='directory to save agent logs')
parser.add_argument(
    '--log_interval',
    type=int,
    default=1,
    help='log interval, one log per n updates')
parser.add_argument(
    "--from_last_checkpoint", 
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help="Begin training from latest checkpoint if available.")
parser.add_argument(
    "--checkpoint_interval", 
    type=int, 
    default=0,
    help="Save model every this many updates.")
parser.add_argument(
    "--archive_interval", 
    type=int, 
    default=0,
    help="Save an archived model every this many updates.")
parser.add_argument(
    "--archive_init_checkpoint", 
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=False,
    help="Archive the initial checkpoint.")
parser.add_argument(
    '--test_interval',
    type=int,
    default=10,
    help='Evaluate on test envs every this many updates.')


# Evaluation args.
eval_parser = parser.add_subparser(
    name='eval',
    prefix='test')
eval_parser.add_argument(
    '--n_episodes',
    type=int,
    default=10,
    help='Number of test episodes per environment')
eval_parser.add_argument(
    '--env_names',
    type=str,
    default=None,
    help='Test environments to evaluate on.')
eval_parser.add_argument(
    '--agent_idxs',
    type=str,
    default='*',
    help="csv of agents to evaluate. '*' indicates all.")
eval_env_parser = parser.add_subparser(
    name='eval_env',
    prefix='test_env',
)

# -------- Maze eval arguments. --------
maze_eval_parser = parser.add_subparser(
    name='maze_eval',
    prefix='maze_test',
    dependency={'env_name': 'Maze*'},
    dest='eval_env'
)
maze_eval_parser.add_argument(
    "--see_agent", 
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=True,
    help="Maze observations include the agent.")
maze_eval_parser.add_argument(
    '--normalize_obs',
    type=str2bool, 
    nargs='?', 
    const=True, 
    default=True,
    help='Ensures observations are between 0 and 1.')


# -------- wandb arguments. --------
wandb_parser = parser.add_subparser(
    name='wandb',
    prefix='wandb')
wandb_parser.add_argument(
    "--base_url",
    type=str,
    default="https://api.wandb.ai",
    help='wandb base url'
)
wandb_parser.add_argument(
    "--api_key",
    type=str,
    default=None,
    help='wandb api key'    
)
wandb_parser.add_argument(
    "--entity",
    type=str,
    default='minqi',
    help='Team name'
)
wandb_parser.add_argument(
    "--project",
    type=str,
    default='paired',
    help='wandb project name for logging'
)
wandb_parser.add_argument(
    "--group",
    type=str,
    default=None,
    help='wandb group name for logging'
)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This file is modified from 
https://github.com/Farama-Foundation/Minigrid

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0
"""

import math

import numpy as np

from minimax.envs.viz.window import Window
import minimax.envs.viz.grid_rendering as rendering
from minimax.envs.maze.maze import OBJECT_TO_INDEX, COLOR_TO_INDEX, COLORS


INDEX_TO_COLOR = [k for k,v in COLOR_TO_INDEX.items()]
TILE_PIXELS = 32


class GridVisualizer:
	"""
	Manages a window and renders contents of EnvState instances to it.
	"""
	tile_cache = {}

	def __init__(self):
		self.window = None

	def _lazy_init_window(self):
		if self.window is None:
			self.window = Window('minimax')

	def show(self, block=False):
		self._lazy_init_window()
		self.window.show(block=block)

	def screenshot(self, path):
		self._lazy_init_window()
		self.window.save_img(path)

	def render(self, params, state, highlight=True, tile_size=TILE_PIXELS, maze_map=None):
		return self._render_state(params, state, highlight, tile_size, maze_map)

	def render_grid(self, grid, tile_size=TILE_PIXELS, k_rot90=0, agent_dir_idx=None):
		self._lazy_init_window()

		img = GridVisualizer._render_grid(
				grid,
				tile_size,
				highlight_mask=None,
				agent_dir_idx=agent_dir_idx,
			)
		
		if k_rot90 > 0:
			img = np.rot90(img, k=k_rot90)

		self.window.show_img(img)

	def _render_state(self, params, state, highlight=True, tile_size=TILE_PIXELS, maze_map=None):
		"""
		Render the state
		"""
		self._lazy_init_window()

		if hasattr(params, 'agent_view_size'):
			agent_view_size = params.agent_view_size
			padding = agent_view_size-2 # show
			grid = np.asarray(state.maze_map[padding:-padding,padding:-padding,:])
		else:
			assert maze_map is not None, 'Either params contains agent_view_size or explicit maze map is passed in.'
			grid = np.asarray(maze_map)

		grid_offset = np.array([1,1])
		h,w = grid.shape[:2]
		
		# === Compute highlight mask
		if highlight:
			highlight_mask = np.zeros(shape=(h,w), dtype=bool)

			f_vec = state.agent_dir
			r_vec = np.array([-f_vec[1], f_vec[0]])

			fwd_bound1 = state.agent_pos
			fwd_bound2 = state.agent_pos + state.agent_dir*(agent_view_size-1)
			side_bound1 = state.agent_pos - r_vec*(agent_view_size//2) 
			side_bound2 = state.agent_pos + r_vec*(agent_view_size//2)

			min_bound = np.min(np.stack([
						fwd_bound1, 
						fwd_bound2, 
						side_bound1, 
						side_bound2]) + grid_offset, 0)

			min_y = min(max(min_bound[1], 0), highlight_mask.shape[0]-1)
			min_x = min(max(min_bound[0],0), highlight_mask.shape[1]-1)

			max_y = min(max(min_bound[1]+agent_view_size, 0), highlight_mask.shape[0]-1)
			max_x = min(max(min_bound[0]+agent_view_size, 0), highlight_mask.shape[1]-1)

			highlight_mask[min_y:max_y,min_x:max_x] = True

		# Render the whole grid
		img = GridVisualizer._render_grid(
			grid,
			tile_size,
			highlight_mask=highlight_mask if highlight else None,
			agent_dir_idx=state.agent_dir_idx if hasattr(state, 'agent_dir_idx') else 0
		)

		self.window.show_img(img)

	@classmethod
	def _render_obj(
		cls,
		obj,
		img):
		# Render each kind of object
		obj_type = obj[0]
		color = INDEX_TO_COLOR[obj[1]]

		if obj_type == OBJECT_TO_INDEX['wall']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['goal']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['agent']:
			agent_dir_idx = obj[2]
			tri_fn = rendering.point_in_triangle(
				(0.12, 0.19),
				(0.87, 0.50),
				(0.12, 0.81),
			)
			tri_fn = rendering.rotate_fn(tri_fn, cx=0.5, cy=0.5, theta=0.5*math.pi*agent_dir_idx)
			rendering.fill_coords(img, tri_fn, (255, 0, 0))
			# rendering.fill_coords(img, tri_fn, (60, 182, 234))

		elif obj_type == OBJECT_TO_INDEX['empty']:
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), COLORS[color])
		elif obj_type == OBJECT_TO_INDEX['lava']:
			c = (255, 128, 0)

			# Background color
			rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 1), c)

			# Little waves
			for i in range(3):
				ylo = 0.3 + 0.2 * i
				yhi = 0.4 + 0.2 * i
				rendering.fill_coords(img, rendering.point_in_line(0.1, ylo, 0.3, yhi, r=0.03), (0,0,0))
				rendering.fill_coords(img, rendering.point_in_line(0.3, yhi, 0.5, ylo, r=0.03), (0,0,0))
				rendering.fill_coords(img, rendering.point_in_line(0.5, ylo, 0.7, yhi, r=0.03), (0,0,0))
				rendering.fill_coords(img, rendering.point_in_line(0.7, yhi, 0.9, ylo, r=0.03), (0,0,0))
		else:
			raise ValueError(f'Rendering object at index {obj_type} is currently unsupported.')

	@classmethod
	def _render_tile(
		cls,
		obj,
		highlight=False,
		agent_dir_idx=None,
		tile_size=TILE_PIXELS,
		subdivs=3
	):
		"""
		Render a tile and cache the result
		"""
		# Hash map lookup key for the cache
		if obj is not None and \
			obj[0] == OBJECT_TO_INDEX['agent'] and \
			agent_dir_idx is not None:
			obj = np.array(obj)
			obj[-1] = agent_dir_idx

		no_object = obj is None or (
			obj[0] in [OBJECT_TO_INDEX['empty'], OBJECT_TO_INDEX['unseen']] \
			and obj[2] == 0
		)

		if not no_object:
			key = (*obj, highlight, tile_size)
		else:
			key = (obj, highlight, tile_size)

		if key in cls.tile_cache:
			return cls.tile_cache[key]

		img = np.zeros(shape=(tile_size * subdivs, tile_size * subdivs, 3), dtype=np.uint8)

		# Draw the grid lines (top and left edges)
		rendering.fill_coords(img, rendering.point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
		rendering.fill_coords(img, rendering.point_in_rect(0, 1, 0, 0.031), (100, 100, 100))

		if not no_object:
			GridVisualizer._render_obj(obj, img)

		if highlight:
			rendering.highlight_img(img)

		# Downsample the image to perform supersampling/anti-aliasing
		img = rendering.downsample(img, subdivs)

		# Cache the rendered tile
		cls.tile_cache[key] = img

		return img

	@classmethod
	def _render_grid(
		cls,
		grid,
		tile_size=TILE_PIXELS,
		highlight_mask=None,
		agent_dir_idx=None):
		if highlight_mask is None:
			highlight_mask = np.zeros(shape=grid.shape[:2], dtype=np.bool)

		# Compute the total grid size in pixels
		width_px = grid.shape[1]*tile_size
		height_px = grid.shape[0]*tile_size

		img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

		# Render the grid
		for y in range(grid.shape[0]):
			for x in range(grid.shape[1]):		
				obj = grid[y,x,:]
				if obj[0] in [OBJECT_TO_INDEX['empty'], OBJECT_TO_INDEX['unseen']] \
					and obj[2] == 0:
					obj = None

				tile_img = GridVisualizer._render_tile(
					obj,
					highlight=highlight_mask[y, x],
					tile_size=tile_size,
					agent_dir_idx=agent_dir_idx,
				)

				ymin = y*tile_size
				ymax = (y+1)*tile_size
				xmin = x*tile_size
				xmax = (x+1)*tile_size
				img[ymin:ymax, xmin:xmax, :] = tile_img

		return img

	def close(self):
		self.window.close()


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from dataclasses import dataclass
from collections import namedtuple, OrderedDict
from functools import partial
from enum import IntEnum

import numpy as np
import jax
import jax.numpy as jnp
from jax import lax
from typing import Tuple, Optional
import chex
from flax import struct
from flax.core.frozen_dict import FrozenDict

from .common import EnvInstance, make_maze_map
from minimax.envs import environment, spaces
from minimax.envs.registration import register_ued


class SequentialActions(IntEnum):
    skip = 0
    wall = 1
    goal = 2
    agent = 3


@struct.dataclass
class EnvState:
	encoding: chex.Array
	time: int
	terminal: bool


@struct.dataclass
class EnvParams:
	height: int = 15
	width: int = 15
	n_walls: int = 25
	noise_dim: int = 50 
	replace_wall_pos: bool = False
	fixed_n_wall_steps: bool = False
	first_wall_pos_sets_budget: bool = False
	use_seq_actions: bool = False,
	set_agent_dir: bool = False
	normalize_obs: bool = False
	singleton_seed: int = -1


class UEDMaze(environment.Environment):
	def __init__(
		self,
		height=13,
		width=13,
		n_walls=25,
		noise_dim=16,
		replace_wall_pos=False,
		fixed_n_wall_steps=False,
		first_wall_pos_sets_budget=False,
		use_seq_actions=False,
		set_agent_dir=False,
		normalize_obs=False,
	):
		"""
		Using the original action space requires ensuring proper handling
		of a sequence with trailing dones, e.g. dones: 0 0 0 0 1 1 1 1 1 ... 1.
		Advantages and value losses should only be computed where ~dones[0].
		"""
		assert not (first_wall_pos_sets_budget and fixed_n_wall_steps), \
			'Setting first_wall_pos_sets_budget=True requires fixed_n_wall_steps=False.'

		super().__init__()

		self.n_tiles = height*width
		self.action_set = jnp.array(jnp.arange(self.n_tiles)) # go straight, turn left, turn right, take action

		self.params = EnvParams(
			height=height,
			width=width,
			n_walls=n_walls,
			noise_dim=noise_dim,
			replace_wall_pos=replace_wall_pos,
			fixed_n_wall_steps=fixed_n_wall_steps,
			first_wall_pos_sets_budget=first_wall_pos_sets_budget,
			use_seq_actions=False,
			set_agent_dir=set_agent_dir,
			normalize_obs=normalize_obs,
		)

	@staticmethod
	def align_kwargs(kwargs, other_kwargs):
		kwargs.update(dict(
			height=other_kwargs['height'],
			width=other_kwargs['width'],
		))

		return kwargs

	def _add_noise_to_obs(self, rng, obs):
		if self.params.noise_dim > 0:
			noise = jax.random.uniform(rng, (self.params.noise_dim,))
			obs.update(dict(noise=noise))

		return obs

	def reset_env(
		self, 
		key: chex.PRNGKey):
		"""
		Prepares the environment state for a new design
		from a blank slate. 
		"""
		params = self.params
		noise_rng, dir_rng = jax.random.split(key)
		encoding = jnp.zeros((self._get_encoding_dim(),), dtype=jnp.uint32)

		if not params.set_agent_dir:
			rand_dir = jax.random.randint(dir_rng, (), minval=0, maxval=4) # deterministic
			tile_scale_dir = jnp.ceil((rand_dir/4)*self.n_tiles).astype(jnp.uint32)
			encoding = encoding.at[-1].set(tile_scale_dir)

		state = EnvState(
			encoding=encoding,
			time=0,
			terminal=False,
		)

		obs = self._add_noise_to_obs(
			noise_rng,
			self.get_obs(state)
		)

		return obs, state

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: int,
	) -> Tuple[chex.Array, EnvState, float, bool, dict]:
		"""
		Take a design step. 
			action: A pos as an int from 0 to (height*width)-1
		"""
		params = self.params

		collision_rng, noise_rng = jax.random.split(key)

		# Sample a random free tile in case of a collision
		dist_values = jnp.logical_and( # True if position taken
			jnp.ones(params.n_walls + 2), 
			jnp.arange(params.n_walls + 2)+1 > state.time
		)

		# Get zero-indexed last wall time step
		if params.fixed_n_wall_steps:
			max_n_walls = params.n_walls
			encoding_pos = state.encoding[:params.n_walls+2]
			last_wall_step_idx = max_n_walls - 1
		else:
			max_n_walls = jnp.round(
				params.n_walls*state.encoding[0]/self.n_tiles).astype(jnp.uint32)

			if self.params.first_wall_pos_sets_budget:
				encoding_pos = state.encoding[:params.n_walls+2]
				last_wall_step_idx = jnp.maximum(max_n_walls,1) - 1
			else:
				encoding_pos = state.encoding[1:params.n_walls+3]
				last_wall_step_idx = max_n_walls

		pos_dist = jnp.ones(self.n_tiles).at[
			jnp.flip(encoding_pos)].set(jnp.flip(dist_values))
		all_pos = jnp.arange(self.n_tiles, dtype=jnp.uint32)

		# Only mark collision if replace_wall_pos=False OR the agent is placed over the goal
		goal_step_idx = last_wall_step_idx + 1
		agent_step_idx = last_wall_step_idx + 2

		# Track whether it is the last time step
		next_state = state.replace(time=state.time + 1)
		done = self.is_terminal(next_state)

		# Always place agent idx in last enc position.
		is_agent_dir_step = jnp.logical_and(
			params.set_agent_dir,
			done
		)

		collision = jnp.logical_and(
			pos_dist[action] < 1,
			jnp.logical_or(
				not params.replace_wall_pos,
				jnp.logical_and( # agent pos cannot override goal
					jnp.equal(state.time, agent_step_idx),
					jnp.equal(state.encoding[goal_step_idx], action)
				)
			)
		)
		collision = (collision * (1-is_agent_dir_step)).astype(jnp.uint32)

		action = (1-collision)*action + \
			collision*jax.random.choice(collision_rng, all_pos, replace=False, p=pos_dist)

		enc_idx = (1-is_agent_dir_step)*state.time + is_agent_dir_step*(-1)
		encoding = state.encoding.at[enc_idx].set(action)

		next_state = next_state.replace(
			encoding=encoding,
			terminal=done
		)
		reward = 0

		obs = self._add_noise_to_obs(noise_rng, self.get_obs(next_state))

		return (
			lax.stop_gradient(obs),
			lax.stop_gradient(next_state),
			reward,
			done,
			{},
		)

	def get_env_instance(
			self, 
			key: chex.PRNGKey,
			state: EnvState
		) -> chex.Array:
		"""
		Converts internal encoding to an instance encoding that 
		can be interpreted by the `set_to_instance` method 
		the paired Environment class.
		"""
		params = self.params
		h = params.height
		w = params.width
		enc = state.encoding

		# === Extract agent_dir, agent_pos, and goal_pos ===
		# Num walls placed currently
		if params.fixed_n_wall_steps:
			n_walls = params.n_walls
			enc_len = self._get_encoding_dim()
			wall_pos_idx = jnp.flip(enc[:params.n_walls])
			agent_pos_idx = enc_len-2 # Enc is full length
			goal_pos_idx = enc_len-3
		else:
			n_walls = jnp.round(
				params.n_walls*enc[0]/self.n_tiles
			).astype(jnp.uint32)
			if params.first_wall_pos_sets_budget:
				wall_pos_idx = jnp.flip(enc[:params.n_walls]) # So 0-padding does not override pos=0
				enc_len = n_walls + 2 # [wall_pos] + len((goal, agent))
			else:
				wall_pos_idx = jnp.flip(enc[1:params.n_walls+1])
				enc_len = n_walls + 3 # [wall_pos] + len((n_walls, goal, agent))
			agent_pos_idx = enc_len-1 # Positions are relative to n_walls when n_walls is variable.
			goal_pos_idx = enc_len-2

		# Get agent + goal info (set agent/goal pos 1-step out of range if they are not yet placed)
		goal_placed = state.time > jnp.array([goal_pos_idx], dtype=jnp.uint32)
		goal_pos = \
			goal_placed*jnp.array([enc[goal_pos_idx]%w, enc[goal_pos_idx]//w], dtype=jnp.uint32) \
			+ (~goal_placed)*jnp.array([w,h], dtype=jnp.uint32)

		agent_placed = state.time > jnp.array([agent_pos_idx], dtype=jnp.uint32)
		agent_pos = \
			agent_placed*jnp.array([enc[agent_pos_idx]%w, enc[agent_pos_idx]//w], dtype=jnp.uint32) \
			+ (~agent_placed)*jnp.array([w,h], dtype=jnp.uint32)

		agent_dir_idx = jnp.floor((4*enc[-1]/self.n_tiles)).astype(jnp.uint8)

		# Make wall map
		wall_start_time = jnp.logical_and( # 1 if explicitly predict # blocks, else 0
			not params.fixed_n_wall_steps,
			not params.first_wall_pos_sets_budget
		).astype(jnp.uint32)
		wall_map = jnp.zeros(h*w, dtype=jnp.bool_)
		wall_values = jnp.arange(params.n_walls) + wall_start_time < jnp.minimum(state.time, n_walls + wall_start_time)
		wall_values = jnp.flip(wall_values)
		wall_map = wall_map.at[wall_pos_idx].set(wall_values)

		# Zero out walls where agent and goal reside
		agent_mask = agent_placed*(~(jnp.arange(h*w) == state.encoding[agent_pos_idx])) + ~agent_placed*wall_map
		goal_mask = goal_placed*(~(jnp.arange(h*w) == state.encoding[goal_pos_idx])) + ~goal_placed*wall_map
		wall_map = wall_map*agent_mask*goal_mask 
		wall_map = wall_map.reshape(h,w)

		return EnvInstance(
			agent_pos=agent_pos,
			agent_dir_idx=agent_dir_idx,
			goal_pos=goal_pos,
			wall_map=wall_map
		)

	def is_terminal(self, state: EnvState) -> bool:
		# if params.fixed_n_wall_steps:
		# 	max_n_walls = params.n_walls
		# 	done_steps = state.time >= self.max_episode_steps()
		# else:
		# 	max_n_walls = jnp.round(params.n_walls*state.encoding[0]/self.n_tiles)
		# 	max_episode_steps = \
		# 		self.max_episode_steps() - (params.n_walls - max_n_walls)
		# 	done_steps = state.time >= max_episode_steps
		done_steps = state.time >= self.max_episode_steps()
		return jnp.logical_or(done_steps, state.terminal)

	def _get_post_terminal_obs(self, state: EnvState):
		dtype = jnp.float32 if self.params.normalize_obs else jnp.uint8
		image = jnp.zeros((
			self.params.height+2, self.params.width+2, 3), dtype=dtype
		)

		return OrderedDict(dict(
			image=image,
			time=state.time,
			noise=jnp.zeros(self.params.noise_dim, dtype=jnp.float32),
		))

	def get_obs(self, state: EnvState):
		instance = self.get_env_instance(jax.random.PRNGKey(0), state)

		image = make_maze_map(
				self.params,
				instance.wall_map, 
				instance.goal_pos, 
				instance.agent_pos, 
				instance.agent_dir_idx,
				pad_obs=False
			)

		if self.params.normalize_obs:
			image = image/10.0

		return OrderedDict(dict(
			image=image,
			time=state.time,
		))

	@property
	def default_params(self):
		return EnvParams()

	@property
	def name(self) -> str:
		"""Environment name."""
		return "UEDMaze"

	@property
	def num_actions(self) -> int:
		"""Number of actions possible in environment."""
		return len(self.action_set)

	def action_space(self) -> spaces.Discrete:
		"""Action space of the environment."""
		params = self.params
		return spaces.Discrete(
			params.height*params.width,
			dtype=jnp.uint32
		)

	def observation_space(self) -> spaces.Dict:
		"""Observation space of the environment."""
		params = self.params
		max_episode_steps = self.max_episode_steps()
		spaces_dict = {
			'image':spaces.Box(0, 255, (params.height+2, params.width+2, 3)),
			'time': spaces.Discrete(max_episode_steps),
		}
		if self.params.noise_dim > 0:
			spaces_dict.update({
				'noise': spaces.Box(0, 1, (self.params.noise_dim,))
			})
		return spaces.Dict(spaces_dict)
			
	def state_space(self) -> spaces.Dict:
		"""State space of the environment."""
		params = self.params
		encoding_dim = self._get_encoding_dim()
		max_episode_steps = self.max_episode_steps()
		h = params.height
		w = params.width
		return spaces.Dict({
			'encoding':spaces.Box(0, 255, (encoding_dim,)),
			'time': spaces.Discrete(max_episode_steps),
			"terminal": spaces.Discrete(2),
		})

	def _get_encoding_dim(self) -> int:
		encoding_dim = self.max_episode_steps()
		if not self.params.set_agent_dir:
			encoding_dim += 1 # max steps is 1 less than full encoding dim

		return encoding_dim

	def max_episode_steps(self) -> int:
		if self.params.fixed_n_wall_steps \
			 or self.params.first_wall_pos_sets_budget:
			max_episode_steps = self.params.n_walls + 2
		else:
			max_episode_steps = self.params.n_walls + 3

		if self.params.set_agent_dir:
			max_episode_steps += 1

		return max_episode_steps


if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register_ued(env_id='Maze', entry_point=module_path + ':UEDMaze')



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from typing import Tuple, Optional

import numpy as np
import jax
import jax.numpy as jnp

import minimax.envs as envs
from minimax.util.rl import (
    AgentPop,
    RollingStats
)
import minimax.util.pytree as _tree_util


def generate_all_kwargs_combos(arg_choices):
    def update_kwargs_with_choices(prev_combos, k, choices):
        updated_combos = []
        for v in choices:
            for p in prev_combos:
                updated = p.copy()
                updated[k] = v
                updated_combos.append(updated)

        return updated_combos

    all_combos = [{}]
    for k, choices in arg_choices.items():
        all_combos = update_kwargs_with_choices(all_combos, k, choices)

    return all_combos


def create_envs_for_kwargs(env_names, kwargs):
    # Check for csv kwargs
    arg_choices = {}
    varied_args = []
    for k,v in kwargs.items():
        if isinstance(v, str) and ',' in v:
            vs = eval(v)
            arg_choices[k] = vs
            varied_args.append(k)
        elif isinstance(v, str):
            arg_choices[k] = [eval(v)]
        else:
            arg_choices[k] = [v]

    # List of kwargs
    kwargs_combos = generate_all_kwargs_combos(arg_choices)

    env_infos = []
    incl_ext = len(varied_args) > 0
    for name in env_names:
        for kwargs in kwargs_combos:
            if incl_ext and len(kwargs) > 0:
                ext = ':'.join([f'{k}={kwargs[k]}' for k in varied_args])
                ext_name = f'{name}:{ext}'
            else:
                ext_name = name
            env_infos.append(
                (name, ext_name, kwargs)
            )

    return env_infos


class EvalRunner:
    def __init__(
        self,
        pop,
        env_names,
        env_kwargs=None,
        n_episodes=10,
        agent_idxs='*',
        render_mode=None):

        self.pop = pop

        if isinstance(agent_idxs, str):
            if "*" in agent_idxs:
                self.agent_idxs = np.arange(pop.n_agents)
            else:
                self.agent_idxs = \
                    np.array([int(x) for x in agent_idxs.split(',')])
        else:
            self.agent_idxs = agent_idxs # assume array

        assert np.max(self.agent_idxs) < pop.n_agents, \
            'Agent index is out of bounds.'

        if isinstance(env_names, str):
            env_names = [
                x.strip() for x in env_names.split(',')
            ]

        self.n_episodes = n_episodes
        env_infos = create_envs_for_kwargs(env_names, env_kwargs)
        env_names = []
        self.ext_env_names = []
        env_kwargs = []
        for (name, ext_name, kwargs) in env_infos:
            env_names.append(name)
            self.ext_env_names.append(ext_name)
            env_kwargs.append(kwargs)
        self.n_envs = len(env_names)

        self.benvs = []
        self.env_params = []
        self.env_has_solved_rate = []
        for env_name, kwargs in zip(env_names, env_kwargs):
            benv = envs.BatchEnv(
                env_name=env_name,
                n_parallel=n_episodes,
                n_eval=1,
                env_kwargs=kwargs,
                wrappers=['monitor_return', 'monitor_ep_metrics']
            )
            self.benvs.append(benv)
            self.env_params.append(benv.env.params)
            self.env_has_solved_rate.append(benv.env.eval_solved_rate is not None)

        self.action_dtype = self.benvs[0].env.action_space().dtype

        monitored_metrics = self.benvs[0].env.get_monitored_metrics()
        self.rolling_stats = RollingStats(names=monitored_metrics, window=1)
        self._update_ep_stats = jax.vmap(
            jax.vmap(
                self.rolling_stats.update_stats, in_axes=(0,0,0,None)),
            in_axes=(0,0,0,None))

        self.test_return_pre = 'test_return'
        self.test_solved_rate_pre = 'test_solved_rate'

        self.render_mode = render_mode
        if render_mode:
            from minimax.envs.viz.grid_viz import GridVisualizer
            self.viz = GridVisualizer()
            self.viz.show()

            if render_mode == 'ipython':
                from IPython import display
                self.ipython_display = display

    def load_checkpoint_state(self, runner_state, state):
        runner_state = list(runner_state)
        runner_state[1] = runner_state[1].load_state_dict(state[1])

        return tuple(runner_state)

    @partial(jax.jit, static_argnums=(0,2))
    def _get_transition(
        self,
        rng,
        benv,
        params,
        state,
        obs,
        carry,
        zero_carry,
        extra):
        value, pi_params, next_carry = self.pop.act(params, obs, carry)
        pi = self.pop.get_action_dist(pi_params, dtype=self.action_dtype)
        rng, subrng = jax.random.split(rng)
        action = pi.sample(seed=subrng)
        log_pi = pi.log_prob(action)

        rng, *vrngs = jax.random.split(rng, self.pop.n_agents+1)

        step_args = (jnp.array(vrngs), state, action, None, extra)
        (next_obs, 
         next_state, 
         reward, 
         done, 
         info, 
         extra) = benv.step(*step_args)

        # Add transition to storage
        step = (obs, action, reward, done, log_pi, value)
        if carry is not None:
            step += (carry,)

        # Zero carry if needed
        if carry is not None:
            next_carry = jax.vmap(_tree_util.pytree_select)(
                done, zero_carry, next_carry)

        if self.render_mode:
            self.viz.render(
                benv.env.params, 
                jax.tree_util.tree_map(lambda x: x[0][0], state))
            if self.render_mode == 'ipython':
                self.ipython_display.display(self.viz.window.fig)
                self.ipython_display.clear_output(wait=True)

        return next_state, next_obs, next_carry, done, info, extra

    @partial(jax.jit, static_argnums=(0, 2))
    def _rollout_benv(
        self, 
        rng, 
        benv,
        params,
        env_params,
        state,
        obs,
        carry,
        zero_carry,
        extra,
        ep_stats):

        def _scan_rollout(scan_carry, rng):
            (state, 
             obs, 
             carry,
             extra, 
             ep_stats) = scan_carry
            
            step = \
                self._get_transition(
                    rng,
                    benv,
                    params, 
                    state, 
                    obs, 
                    carry, 
                    zero_carry,
                    extra)

            (next_state, 
             next_obs, 
             next_carry, 
             done, 
             info, 
             extra) = step

            ep_stats = self._update_ep_stats(ep_stats, done, info, 1)

            return (next_state, next_obs, next_carry, extra, ep_stats), None

        n_steps = benv.env.max_episode_steps()
        rngs = jax.random.split(rng, n_steps)
        (state, 
         obs, 
         carry, 
         extra,
         ep_stats),_ = jax.lax.scan(
            _scan_rollout,
            (state, obs, carry, extra, ep_stats),
            rngs,
            length=n_steps)

        return ep_stats

    @partial(jax.jit, static_argnums=(0,))
    def run(self, rng, params):
        """
        Rollout agents on each env. 

        For each env, run n_eval episodes in parallel, 
        where each is indexed to return in order.
        """
        eval_stats = self.fake_run(rng, params)
        rng, *rollout_rngs = jax.random.split(rng, self.n_envs+1)
        for i, (benv, env_param) in enumerate(zip(self.benvs, self.env_params)):
            rng, *reset_rngs = jax.random.split(rng, self.pop.n_agents+1)
            obs, state, extra = benv.reset(jnp.array(reset_rngs))

            if self.pop.agent.is_recurrent:
                rng, subrng = jax.random.split(rng)
                zero_carry = self.pop.init_carry(subrng, obs)
            else:
                zero_carry = None

            # Reset episodic stats
            ep_stats = self.rolling_stats.reset_stats(
                batch_shape=(self.pop.n_agents, self.n_episodes))

            ep_stats = self._rollout_benv(
                rollout_rngs[i],
                benv,
                jax.lax.stop_gradient(params), 
                env_param, 
                state, 
                obs,
                zero_carry,
                zero_carry,
                extra,
                ep_stats)
            
            env_name = self.ext_env_names[i]
            mean_return = ep_stats['return'].mean(1)

            if self.env_has_solved_rate[i]:
                mean_solved_rate = jax.vmap(jax.vmap(benv.env.eval_solved_rate))(ep_stats).mean(1)

            for idx in self.agent_idxs:
                eval_stats[f'eval/a{idx}:{self.test_return_pre}:{env_name}'] = mean_return[idx].squeeze()
                if self.env_has_solved_rate[i]:
                    eval_stats[f'eval/a{idx}:{self.test_solved_rate_pre}:{env_name}'] = mean_solved_rate[idx].squeeze()

        return eval_stats

    def fake_run(self, rng, params):
        eval_stats = {}
        for i, env_name in enumerate(self.ext_env_names):
            for idx in self.agent_idxs:
                eval_stats.update({
                    f'eval/a{idx}:{self.test_return_pre}:{env_name}':0.
                })
                if self.env_has_solved_rate[i]:
                    eval_stats.update({
                        f'eval/a{idx}:{self.test_solved_rate_pre}:{env_name}':0.,
                    })

        return eval_stats



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import importlib

from .wrappers import *
from minimax.envs.environment_ued import UEDEnvironment


# Global registry
registered_envs = []

env2entry = {}

env2ued_entry = {}

env2comparator_entry = {}

env2mutator_entry = {}

name2wrapper = {
	'env_wrapper': EnvWrapper, # for testing,
	'ued_env_wrapper': UEDEnvWrapper,
	'monitor_return': MonitorReturnWrapper,
	'monitor_ep_metrics': MonitorEpisodicMetricsWrapper,
}


def _load(name):
	mod_name, attr_name = name.split(":")
	mod = importlib.import_module(mod_name)
	fn = getattr(mod, attr_name)
	return fn


def _fn_for_entry(entry):
	if callable(entry):
		return entry
	else:
		return  _load(entry)


def cls_for_env_id(env_id):
	if env_id not in env2entry.keys():
		raise ValueError(f"{env_id} is not registered.")
	else:
		entry = env2entry[env_id]
		return _fn_for_entry(entry)


def _make_env(entry, **env_kwargs):
	return _fn_for_entry(entry)(**env_kwargs)


def _apply_wrappers(env, wrappers):
	base_env = env
	if wrappers is not None and len(wrappers) > 0:
		for name in wrappers:
			wrapper_cls = name2wrapper[name]
			if wrapper_cls.is_compatible(base_env):
				env = wrapper_cls(env)

	return env


def make(
	env_id: str, 
	env_kwargs={}, 
	ued_env_kwargs={}, 
	wrappers=None,
	ued_wrappers=None
):  
	"""The minimax equivalent of OpenAI's env.make(env_name)"""
	if env_id not in env2entry.keys():
		raise ValueError(f"{env_id} is not registered.")
	else:
		entry = env2entry[env_id]
		env = _make_env(entry, **env_kwargs)

	if len(ued_env_kwargs) > 0:
		if env_id not in env2ued_entry.keys():
			raise ValueError(f"{env_id} has no UED counterpart registered.")

		_env_kwargs = env_kwargs
		env_kwargs = env.default_params.__dict__
		env_kwargs.update(_env_kwargs)

		ued_entry = env2ued_entry[env_id]
		ued_env_kwargs = _fn_for_entry(ued_entry).align_kwargs(ued_env_kwargs, env_kwargs)

		ued_env = _make_env(ued_entry, **ued_env_kwargs)

		env = UEDEnvironment(env=env, ued_env=ued_env)

	base_env = env

	env = _apply_wrappers(env, wrappers)

	if isinstance(base_env, UEDEnvironment):
		env = _apply_wrappers(env, ued_wrappers)
		return env, env.env.params, env.ued_env.params

	return env, env.params


def get_comparator(env_id: str, comparator_id: str = 'default'):
	entry_point = env2comparator_entry[env_id].get(comparator_id, None)
	assert entry_point is not None, f'Undefined comparator {comparator_id} for environment {env_id}.'

	return _fn_for_entry(entry_point)


def get_mutator(env_id: str, mutator_id: str = 'default'):
	entry_point = env2mutator_entry[env_id].get(mutator_id, None)
	assert entry_point is not None, f'Undefined mutator {mutator_id} for environment {env_id}.'

	return _fn_for_entry(entry_point)


def register(env_id: str, entry_point: str):
	env2entry[env_id] = entry_point


def register_ued(env_id: str, entry_point: str):
	env2ued_entry[env_id] = entry_point


def register_comparator(env_id: str, entry_point: str, comparator_id: str = None):
	if comparator_id is None:
		comparator_id = 'default'

	if env_id not in env2comparator_entry:
		env2comparator_entry[env_id] = {}
	env2comparator_entry[env_id][comparator_id] = entry_point


def register_mutator(env_id: str, entry_point: str, mutator_id: str = None):
	if mutator_id is None:
		mutator_id = 'default'

	if env_id not in env2mutator_entry:
		env2mutator_entry[env_id] = {}
	env2mutator_entry[env_id][mutator_id] = entry_point



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from collections import OrderedDict
from typing import Tuple, Optional

import jax
from flax import struct
import chex

from envs import environment
from envs import spaces
from envs.registration import register, register_ued


@struct.dataclass
class EnvState:
	time: int = 0
	terminal: bool = False

@struct.dataclass
class EnvParams:
	reward_per_step: int = 1.0
	max_episode_steps: int = 250


class DummyRewardEnv(environment.Environment):
	def __init__(self, reward_per_step=1.0, max_episode_steps=250):
		self.reward_per_step = 1.0

		self.params = EnvParams(
			reward_per_step=reward_per_step,
			max_episode_steps=max_episode_steps
		)

	@staticmethod
	def align_kwargs(kwargs, other_kwargs):
		return kwargs

	def reset_env(
		self, 
		key: chex.PRNGKey, 
	) -> Tuple[chex.ArrayTree, EnvState]:
		state = EnvState(
			time=0,
			terminal=False
		)

		return self.get_obs(state), state

	def step_env(
		self,
		key: chex.PRNGKey,
		state: EnvState,
		action: int,
	) -> Tuple[chex.Array, EnvState, float, bool, dict]:
		next_time = state.time + 1
		done = next_time >= self.params.max_episode_steps

		state = state.replace(
			time=next_time,
			terminal=done
		)

		return (
			jax.lax.stop_gradient(self.get_obs(state)),
			jax.lax.stop_gradient(state),
			self.params.reward_per_step,
			done,
			{},
		)

	def get_obs(self, state: EnvState) -> chex.ArrayTree:
		return OrderedDict(dict(time=state.time))

	@property
	def default_params(self) -> EnvParams:
		return EnvParams()

	@property
	def name(self) -> str:
		return "DummyRewardEnv"

	@property
	def num_actions(self) -> int:
		return len(self.action_set)

	def action_space(self) -> spaces.Discrete:
		return spaces.Discrete(1)

	def observation_space(self) -> spaces.Dict:
		return spaces.Dict({
			"time": spaces.Discrete(self.params.max_episode_steps),
		})

	def state_space(self) -> spaces.Dict:
		return spaces.Dict({
			"time": spaces.Discrete(self.params.max_episode_steps),
			"terminal": spaces.Discrete(2),
		})

	def max_episode_steps(self) -> int:
		return self.params.max_episode_steps

	# UED-specific
	def get_env_instance(
			self, 
			key: chex.PRNGKey,
			state: EnvState
		) -> chex.ArrayTree:
		return state

	def set_env_instance(self, encoding: chex.ArrayTree):
		state = encoding
		return self.get_obs(state), state


# Register the env as its own UED env
if hasattr(__loader__, 'name'):
  module_path = __loader__.name
elif hasattr(__loader__, 'fullname'):
  module_path = __loader__.fullname

register(env_id='DummyRewardEnv', entry_point=module_path + ':DummyRewardEnv')
register_ued(env_id='DummyRewardEnv', entry_point=module_path + ':DummyRewardEnv')



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from .maze import Maze
from .maze_ued import UEDMaze
from .maze_ood import *

from .maze_comparators import *
from .maze_mutators import *


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import copy


class DotDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, dct):
        for key, value in dct.items():
            if hasattr(value, 'keys'):
                value = DotDict(value)
            self[key] = value

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self

    def __deepcopy__(self, memo):
        return DotDict(copy.deepcopy(dict(self)))


class DefaultDotDict(dict):
    __delattr__ = dict.__delitem__

    def __init__(self, dct, default=None):
        super().__init__(dct)
        self._default = default

    def __getstate__(self):
        return (self, self._default)

    def __setstate__(self, state):
        self.update(state[0])
        self.__dict__ = self
        self._default = state[1]

    def __setattr__(self, name, value):
        if name == '_default':
            super().__setattr__('_default', value)
        else:
            self.__setitem__(name, value)

    def __deepcopy__(self, memo):
        return DefaultDotDict(
            copy.deepcopy(dict(self)), 
            default=self._default
        )

    def __getattr__(self, name):
        if name == '_default':
            return self._default
        else:
            try:
                return self.__getitem__(name)
            except:
                return self._default


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial
from typing import Tuple, Optional
import inspect

import numpy as np
import jax
import jax.numpy as jnp
from jax.sharding import PartitionSpec as P
import optax
import flax
import flax.linen as nn
from flax.core.frozen_dict import FrozenDict

import minimax.envs as envs
from minimax.util import pytree as _tree_util
from minimax.util.rl import (
	AgentPop,
	VmapTrainState,
	RolloutStorage,
	RollingStats
)


class DRRunner:
	"""
	Orchestrates rollouts across one or more students. 
	The main components at play:
	- AgentPop: Manages train state and batched inference logic 
		for a population of agents.
	- BatchEnv: Manages environment step and reset logic, using a 
		populaton of agents.
	- RolloutStorage: Manages the storing and sampling of collected txns.
	- PPO: Handles PPO updates, which take a train state + batch of txns.
	"""
	def __init__(
		self, 
		env_name,
		env_kwargs,
		student_agents,
		n_students=1,
		n_parallel=1,
		n_eval=1,
		n_rollout_steps=256,
		lr=1e-4,
		lr_final=None,
		lr_anneal_steps=0,
		max_grad_norm=0.5,
		discount=0.99,
		gae_lambda=0.95,
		adam_eps=1e-5,
		normalize_return=False,
		track_env_metrics=False,
		n_unroll_rollout=1,
		n_devices=1,
		render=False):

		assert len(student_agents) == 1, 'Only one type of student supported.'
		assert n_parallel % n_devices == 0, 'Num envs must be divisible by num devices.'

		self.n_students = n_students
		self.n_parallel = n_parallel // n_devices
		self.n_eval = n_eval
		self.n_devices = n_devices
		self.step_batch_size = n_students*n_eval*n_parallel
		self.n_rollout_steps = n_rollout_steps
		self.n_updates = 0
		self.lr = lr
		self.lr_final = lr if lr_final is None else lr_final
		self.lr_anneal_steps = lr_anneal_steps
		self.max_grad_norm = max_grad_norm
		self.adam_eps = adam_eps
		self.normalize_return = normalize_return
		self.track_env_metrics = track_env_metrics
		self.n_unroll_rollout = n_unroll_rollout
		self.render = render

		self.student_pop = AgentPop(student_agents[0], n_agents=n_students)

		self.env, self.env_params = envs.make(
			env_name, 
			env_kwargs=env_kwargs
		)
		self._action_shape = self.env.action_space().shape

		self.benv = envs.BatchEnv(
			env_name=env_name,
			n_parallel=self.n_parallel,
			n_eval=self.n_eval,
			env_kwargs=env_kwargs,
			wrappers=['monitor_return', 'monitor_ep_metrics']
		)
		self.action_dtype = self.benv.env.action_space().dtype

		self.student_rollout = RolloutStorage(
			discount=discount,
			gae_lambda=gae_lambda,
			n_steps=n_rollout_steps,
			n_agents=n_students,
			n_envs=self.n_parallel,
			n_eval=self.n_eval,
			action_space=self.env.action_space(),
			obs_space=self.env.observation_space(),
			agent=self.student_pop.agent,
		)

		monitored_metrics = self.benv.env.get_monitored_metrics()
		self.rolling_stats = RollingStats(
			names=monitored_metrics,
			window=10,
		)
		self._update_ep_stats = jax.vmap(jax.vmap(self.rolling_stats.update_stats))

		if self.render:
			from envs.viz.grid_viz import GridVisualizer
			self.viz = GridVisualizer()
			self.viz.show()

	def reset(self, rng):
		self.n_updates = 0

		n_parallel = self.n_parallel*self.n_devices

		rngs, *vrngs = jax.random.split(rng, self.n_students+1)
		obs, state, extra = self.benv.reset(jnp.array(vrngs), n_parallel=n_parallel)
		dummy_obs = jax.tree_util.tree_map(lambda x: x[0], obs) # for one agent only

		rng, subrng = jax.random.split(rng)
		if self.student_pop.agent.is_recurrent:
			carry = self.student_pop.init_carry(subrng, obs)
			self.zero_carry = jax.tree_map(lambda x: x.at[:,:self.n_parallel].get(), carry)
		else:
			carry = None

		rng, subrng = jax.random.split(rng)
		params = self.student_pop.init_params(subrng, dummy_obs)

		schedule_fn = optax.linear_schedule(
			init_value=-float(self.lr),
			end_value=-float(self.lr_final),
			transition_steps=self.lr_anneal_steps,
		)

		tx = optax.chain(
			optax.clip_by_global_norm(self.max_grad_norm),
			optax.adam(learning_rate=float(self.lr), eps=self.adam_eps)
		)

		train_state = VmapTrainState.create(
			apply_fn=self.student_pop.agent.evaluate,
			params=params,
			tx=tx
		)

		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=(self.n_students, n_parallel*self.n_eval))

		start_state = state

		return (
			rng, 
			train_state, 
			state,
			start_state, # Used to track metrics from starting state
			obs, 
			carry, 
			extra, 
			ep_stats
		)

	def get_checkpoint_state(self, state):
		_state = list(state)
		_state[1] = state[1].state_dict

		return _state

	def load_checkpoint_state(self, runner_state, state):
		runner_state = list(runner_state)
		runner_state[1] = runner_state[1].load_state_dict(state[1])

		return tuple(runner_state)

	@partial(jax.jit, static_argnums=(0,2))
	def _get_transition(
		self, 
		rng, 
		pop, 
		params, 
		rollout, 
		state, 
		start_state, 
		obs, 
		carry, 
		done,
		reset_state=None,
		extra=None):
		# Sample action
		value, pi_params, next_carry = pop.act(params, obs, carry, done)

		pi = pop.get_action_dist(pi_params, dtype=self.action_dtype)
		rng, subrng = jax.random.split(rng)
		action = pi.sample(seed=subrng)
		log_pi = pi.log_prob(action)

		rng, *vrngs = jax.random.split(rng, self.n_students+1)
		(next_obs, 
		 next_state, 
		 reward, 
		 done, 
		 info, 
		 extra) = self.benv.step(jnp.array(vrngs), state, action, reset_state, extra)

		next_start_state = jax.vmap(_tree_util.pytree_select)(
			done, next_state, start_state
		)

		# Add transition to storage
		step = (obs, action, reward, done, log_pi, value)
		if carry is not None:
			step += (carry,)

		rollout = self.student_rollout.append(rollout, *step)

		if self.render:
			self.viz.render(
				self.benv.env.params, 
				jax.tree_util.tree_map(lambda x: x[0][0], state))

		return (
			rollout, 
			next_state,
			next_start_state, 
			next_obs, 
			next_carry, 
			done, 
			info, 
			extra
		)

	@partial(jax.jit, static_argnums=(0,))
	def _rollout_students(
		self, 
		rng, 
		train_state, 
		state, 
		start_state, 
		obs, 
		carry, 
		done,
		reset_state=None,
		extra=None, 
		ep_stats=None):
		rollout = self.student_rollout.reset()

		rngs = jax.random.split(rng, self.n_rollout_steps)

		def _scan_rollout(scan_carry, rng):
			rollout, state, start_state, obs, carry, done, extra, ep_stats, train_state = scan_carry 

			next_scan_carry = \
				self._get_transition(
					rng, 
					self.student_pop, 
					jax.lax.stop_gradient(train_state.params), 
					rollout, 
					state,
					start_state, 
					obs, 
					carry,
					done,
					reset_state, 
					extra)
			(rollout, 
			 next_state,
			 next_start_state, 
			 next_obs, 
			 next_carry, 
			 done, 
			 info, 
			 extra) = next_scan_carry

			ep_stats = self._update_ep_stats(ep_stats, done, info)

			return (
				rollout, 
				next_state,
				next_start_state,
				next_obs, 
				next_carry,
				done,
				extra, 
				ep_stats,
				train_state), None

		(rollout, 
		 state, 
		 start_state, 
		 obs, 
		 carry, 
		 done,
		 extra, 
		 ep_stats,
		 train_state), _ = jax.lax.scan(
			_scan_rollout,
			(rollout, 
			 state, 
			 start_state,
			 obs, 
			 carry, 
			 done,
			 extra, 
			 ep_stats,
			 train_state),
			rngs,
			length=self.n_rollout_steps,
		)

		return rollout, state, start_state, obs, carry, extra, ep_stats, train_state

	@partial(jax.jit, static_argnums=(0,4))
	def _compile_stats(self, update_stats, ep_stats, env_metrics=None, mask_passable=True):
		stats = jax.vmap(lambda info: jax.tree_map(lambda x: x.mean(), info))(
			{k:ep_stats[k] for k in self.rolling_stats.names}
		)
		stats.update(update_stats)

		if self.n_students > 1:
			_stats = {}
			for i in range(self.n_students):
				_student_stats = jax.tree_util.tree_map(lambda x: x[i], stats) # for agent0
				_stats.update({f'a{i}/{k}':v for k,v in _student_stats.items()})
			stats = _stats

		if self.track_env_metrics:
			if mask_passable:
				passable_mask = env_metrics.pop('passable')
				mean_env_metrics = jax.vmap(lambda info, mask: jax.tree.map(lambda x: (x*mask).sum()/mask.sum(), info))(env_metrics, passable_mask)
				mean_env_metrics = {f'env/{k}':v for k,v in mean_env_metrics.items()}
				mean_env_metrics.update({'env/passable_ratio': jax.vmap(lambda x: x.mean())(passable_mask)})
			else:
				mean_env_metrics = jax.vmap(lambda info: jax.tree.map(lambda x: x.mean(), info))(env_metrics)
				mean_env_metrics = {f'env/{k}':v for k,v in mean_env_metrics.items()}

			if self.n_students > 1:
				_env_metrics = {}
				for i in range(self.n_students):
					_student_env_metrics = jax.tree_util.tree_map(lambda x: x[i], mean_env_metrics) # for agent0
					_env_metrics.update({f'{k}_a{i}':v for k,v in _student_env_metrics.items()})
				mean_env_metrics = _env_metrics

			stats.update(mean_env_metrics)

		if self.n_students == 1:
			stats = jax.tree_map(lambda x: x[0], stats)

		if self.n_devices > 1:
			stats = jax.tree_map(lambda x: jax.lax.pmean(x, 'device'), stats)

		return stats

	def get_shmap_spec(self):
		runner_state_size = len(inspect.signature(self.run).parameters)
		in_spec = [P(None,'device'),]*(runner_state_size)
		out_spec = [P(None,'device'),]*(runner_state_size)

		in_spec[:2] = [P(None),]*2
		in_spec = tuple(in_spec)
		out_spec = (P(None),) + in_spec

		return in_spec, out_spec

	@partial(jax.jit, static_argnums=(0,))
	def run(
		self, 
		rng, 
		train_state, 
		state, 
		start_state,
		obs, 
		carry=None, 
		extra=None, 
		ep_stats=None):
		"""
		Perform one update step: rollout all students and teachers + update with PPO
		"""
		if self.n_devices > 1:
			rng = jax.random.fold_in(rng, jax.lax.axis_index('device'))

		rng, *vrngs = jax.random.split(rng, self.n_students+1)
		rollout_batch_shape = (self.n_students, self.n_parallel*self.n_eval)

		obs, state, extra = self.benv.reset(jnp.array(vrngs))
		ep_stats = self.rolling_stats.reset_stats(
			batch_shape=rollout_batch_shape)

		rollout_start_state = state

		done = jnp.zeros(rollout_batch_shape, dtype=jnp.bool_)
		reset_state = None  # Sample new environment on reset
		rng, subrng = jax.random.split(rng)
		rollout, state, start_state, obs, carry, extra, ep_stats, train_state = \
			self._rollout_students(
				subrng, 
				train_state, 
				state, 
				start_state,
				obs, 
				carry, 
				done,
				reset_state,
				extra,
				ep_stats
			)

		train_batch = self.student_rollout.get_batch(
			rollout, 
			self.student_pop.get_value(
				jax.lax.stop_gradient(train_state.params), 
				obs, 
				carry,
			)
		)

		# PPOAgent vmaps over the train state and batch. Batch must be N x EM
		rng, subrng = jax.random.split(rng)
		train_state, update_stats = self.student_pop.update(subrng, train_state, train_batch)

		# Collect env metrics
		if self.track_env_metrics:
			env_metrics = self.benv.get_env_metrics(rollout_start_state)
		else:
			env_metrics = None

		stats = self._compile_stats(update_stats, ep_stats, env_metrics)
		stats.update(dict(n_updates=train_state.n_updates[0]))

		train_state = train_state.increment()
		self.n_updates += 1

		return (
			stats, 
			rng, 
			train_state, 
			state, 
			start_state, 
			obs, 
			carry, 
			extra, 
			ep_stats
		)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import collections

import jax


def pytree_set_array_at(pytree, i, value):
	return jax.tree_util.tree_map(lambda x,y: x.at[i].set(y), pytree, value)

def pytree_set_struct_at(pytree, i, value):
	return jax.tree_util.tree_map(lambda x,y: x.at[i].set(y), pytree, value)

def pytree_at(pytree, start, end=None):
	return jax.tree_util.tree_map(lambda x: x.at[start:end].get(), pytree)

def pytree_select(pred, on_true, on_false):
	vselect = jax.vmap(jax.lax.select, in_axes=(0, 0, 0))
	return jax.tree_util.tree_map(lambda x,y: vselect(pred, x, y), on_true, on_false)

def pytree_expand_batch_dim(pytree, batch_shape, n_batch_axes=2):
	"""
	Expands a single batch dimension into a multi-dim batch shape 
	"""
	return jax.tree_util.tree_map(lambda x: x.reshape(*batch_shape, *x.shape[n_batch_axes:]), pytree)

def pytree_transform(pytree, transform):
	return jax.tree_util.tree_map(lambda x: transform(x), pytree)

def pytree_merge(dst, src, start_idx, src_len):
	return jax.tree_map(lambda x,y: x.at[start_idx:start_idx+src_len].set(y.at[:src_len].get()), dst, src)


"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

from functools import partial

import jax
import jax.numpy as jnp


class RollingStats:
	"""
	This class tracks episodic stats, such as final returns 
	and env complexity metrics. Works on a per-env basis.
	"""
	def __init__(self, names, step_metrics_names=[], window=None):
		self.names = names
		self.step_metric_names = step_metrics_names
		self.window = window

	@partial(jax.jit, static_argnums=(0,1))
	def reset_stats(self, batch_shape=(1,)):
		stats = {
			'n_episodes': jnp.zeros((*(batch_shape),1), dtype=jnp.uint32),
			'n_steps': jnp.zeros((*(batch_shape),1), dtype=jnp.uint32)
		}
		stats.update({
			name: jnp.zeros((*(batch_shape),1)) for name in self.names
		})

		if self.window is not None:
			# Average over window      
			stats.update({
				f'{name}_buffer': jnp.zeros((*(batch_shape), self.window)) \
					for name in self.names
			})

		return stats

	@partial(jax.jit, static_argnums=(0,))
	def update_stats(self, stats, done, info, max_episodes=jnp.inf):
		n_eps = stats['n_episodes']
		n_steps = stats['n_steps']

		for name in self.names:
			# Update stat
			if name not in info:
				continue

			new_val = info[name]

			# Only record first max_episode episodes
			done = done*(n_eps < max_episodes)
			if name in self.step_metric_names:
				n_incr_prev = n_steps
				n_incr_total = n_steps + 1
				_metric_done = True
			else:
				n_incr_prev = n_eps
				n_incr_total = n_eps + done
				_metric_done = done

			if self.window is None:
				mean = stats[name]
				new_mean = self._update_stat_mean(
					new_val, mean, n_incr_total, _metric_done 
				)
			else:
				buffer_key = f'{name}_buffer'
				buffer = stats[buffer_key]
				new_mean, buffer = self._update_stat_window(
					new_val, buffer, n_incr_total, n_incr_prev, _metric_done
				)
				stats.update({buffer_key:buffer})

			stats.update({
				name: new_mean,
			})

			# Only update n_episodes based on real episodes
			if name in self.step_metric_names:
				stats.update({
					'n_steps': n_incr_total
				})
			else:
				stats.update({
					'n_episodes': n_incr_total
				})

		return stats

	@partial(jax.jit, static_argnums=(0,))
	def _update_stat_mean(self, new_val, mean, n_eps_total, done):
		z = 1/jnp.maximum(1, n_eps_total)
		new_mean = done*(mean*(1-z) + new_val*z) + (1-done)*mean

		return new_mean

	@partial(jax.jit, static_argnums=(0,))
	def _update_stat_window(self, new_val, buffer, n_eps_total, n_eps_prev, done):
		cur_val = buffer[n_eps_prev % self.window]
		new_val = done*new_val + (1-done)*cur_val
		buffer = buffer.at[n_eps_prev % self.window].set(new_val)
		new_mean = buffer.sum()/jnp.maximum(jnp.minimum(self.window, n_eps_total),1)

		return new_mean, buffer



"""
Copyright (c) Meta Platforms, Inc. and affiliates.
All rights reserved.

This source code is licensed under the license found in the
LICENSE file in the root directory of this source tree.
"""

import pytest
import numpy as np
import jax
import jax.numpy as jnp

from tests.base_req_rollout_storage import RequiresRolloutStorageTestClass


class TestRolloutStorage(RequiresRolloutStorageTestClass):
	def test_reset(self):
		rollout = self.rollout_mgr.reset()

		batch_shape = self.batch_shape
		obs_space = self.env.observation_space()
		for k,v in rollout['obs'].items():
			assert rollout['obs'][k].shape == (
				*batch_shape,
				*obs_space.spaces[k].shape
			)
			assert (rollout['obs'][k] > 0).sum() == 0

		assert rollout['actions'].shape == (
			*batch_shape,
			*self.env.action_space().shape
		)

		assert rollout['rewards'].shape == batch_shape
		assert (rollout['rewards'] > 0).sum() == 0

		assert rollout['dones'].shape == batch_shape
		assert (rollout['dones'] > 0).sum() == 0

		assert rollout['log_pis_old'].shape == batch_shape
		assert (rollout['log_pis_old'] > 0).sum() == 0

		assert rollout['values_old'].shape == batch_shape
		assert (rollout['values_old'] > 0).sum() == 0

		assert rollout['_t'].shape == (self.n_agents,)
		assert (rollout['_t'] > 0).sum() == 0

		assert rollout['carry'][0].shape == (*batch_shape, self.rnn_dim)
		assert rollout['carry'][1].shape == (*batch_shape, self.rnn_dim)
		assert jax.tree_util.tree_structure(rollout['carry'][0].at[0,0,0].get()) \
			== jax.tree_util.tree_structure(self.zero_carry_t[0])
		assert jax.tree_util.tree_structure(rollout['carry'][1].at[0,0,0].get()) \
			== jax.tree_util.tree_structure(self.zero_carry_t[1])
		assert (rollout['carry'][0] > 0).sum() == 0
		assert (rollout['carry'][1] > 0).sum() == 0

	def test_append(self):
		# Make sure appending the full rollout length looks right
		n_appends = 10

		t_batch_shape = self.t_batch_shape

		rollout = self.rollout_mgr.reset()

		step = self._get_dummy_step_components()

		for i in range(n_appends):
			rollout = self.rollout_mgr.append(rollout, *step)

		t_batch_size = np.prod(t_batch_shape)

		assert rollout['actions'].sum() == t_batch_size*n_appends
		assert rollout['rewards'].sum() == t_batch_size*n_appends
		assert rollout['dones'].sum() == t_batch_size*n_appends
		assert rollout['log_pis_old'].sum() == t_batch_size*n_appends
		assert rollout['values_old'].sum() == t_batch_size*n_appends
		assert rollout['_t'].mean() == n_appends

		t_overshoot = 2
		for t in range(self.n_steps - n_appends + t_overshoot):
			rollout = self.rollout_mgr.append(rollout, *step)

		assert rollout['_t'].mean() == t_overshoot

	def test_compute_gae(self):
		# Set up placeholder values
		(obs, 
		 action, 
		 _, 
		 _, 
		 log_pi, 
		 _, 
		 carry) = self._get_dummy_step_components()

		# Mark episode done every 8 steps
		batch_shape = self.batch_shape
		done = jnp.zeros(batch_shape, dtype=jnp.uint8)
		done = done.at[:,jnp.arange(4,self.n_steps,4),:].set(1)

		# Reward of 10 at the end of every episode
		reward = done*10

		# Predict 0.1 at every time step
		value = jnp.ones(batch_shape)*0.1

		# Last value is 1
		last_value = jnp.ones(self.t_batch_shape)

		advantages, targets = jax.vmap(self.rollout_mgr.compute_gae)(
			value, reward, done, last_value
		)

		_advantages = jnp.zeros(batch_shape)

		next_value = last_value
		next_advantage = np.zeros_like(advantages.at[:,0,:].get())
		for t in np.arange(self.n_steps)[::-1]:
			_done = done.at[:,t,:].get()
			cur_value = value.at[:,t,:].get()
			td = reward.at[:,t,:].get() + self.discount*next_value*(1-_done) - cur_value
			_advantages = \
				_advantages.at[:,t,:].set(td + self.discount*self.gae_lambda*(1-_done)*next_advantage)
			next_advantage = _advantages.at[:,t,:].get()
			next_value = cur_value

		_targets = _advantages + value

		assert (_advantages != advantages).sum() == 0
		assert (_targets != targets).sum() == 0

	def test_get_batch(self):
		rollout = self.rollout_mgr.reset()
		step = self._get_dummy_step_components()
		for i in range(self.n_steps):
			rollout = self.rollout_mgr.append(rollout, *step)

		last_value = jnp.ones(self.t_batch_shape)

		batch = self.rollout_mgr.get_batch(rollout, last_value)

		for k,v in batch.obs.items():
			assert (batch.obs[k] != rollout['obs'][k]).sum() == 0

		assert (batch.actions != rollout['actions']).sum() == 0
		assert (batch.dones != rollout['dones']).sum() == 0
		assert (batch.rewards != rollout['rewards']).sum() == 0
		assert (batch.log_pis != rollout['log_pis_old']).sum() == 0
		assert (batch.values != rollout['values_old']).sum() == 0

		assert ((batch.advantages + batch.values) != batch.targets).sum() == 0

		assert (batch.carry[0] != rollout['carry'][0]).sum() == 0
		assert (batch.carry[1] != rollout['carry'][1]).sum() == 0

	def test_set_final_reward(self):
		rollout = self.rollout_mgr.reset()

		final_reward = jnp.ones(self.t_batch_shape)*3
		rollout = self.rollout_mgr.set_final_reward(rollout, final_reward)

		assert (rollout['rewards'].at[:,-1,:].get() != final_reward).sum() == 0
